from __future__ import division
from __future__ import print_function

'''
Test implementation using cell models of the Blue Brain Project with LFPy.
The example assumes that cell models available from
https://bbpnmc.epfl.ch/nmc-portal/downloads are unzipped in the folder 'cell_models'

The function compile_all_mechanisms must be run once before any cell simulation
'''

import os
from os.path import join
import sys
from glob import glob
import numpy as np
import MEAutility as mu
import yaml
import time
from distutils.version import StrictVersion

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False


def get_templatename(f):
    '''
    Assess from hoc file the templatename being specified within

    Arguments
    ---------
    f : file, mode 'r'

    Returns
    -------
    templatename : str

    '''
    templatename = None
    f = open("template.hoc", 'r')
    for line in f.readlines():
        if 'begintemplate' in line.split():
            templatename = line.split()[-1]
            continue
    return templatename


def compile_all_mechanisms(cell_folder, verbose=False):
    """ Attempt to set up a folder with all unique mechanism *.mod files and 
        compile them all. assumes all cell models are in a folder 'cell_models'
    
    Parameters:
    -----------
    model : string (optional, default='bbp')
        Cell model type ('bbp' - Blue Brain Project, i.e. NMC database)
    """

    if not os.path.isdir(join(cell_folder, 'mods')):
        os.mkdir(join(cell_folder, 'mods'))

    neurons = [join(cell_folder, f) \
               for f in os.listdir(join(cell_folder)) \
               if f != 'mods']
    if verbose:
        print(neurons)

    for nrn in neurons:
        for nmodl in glob(join(nrn, 'mechanisms', '*.mod')):
            while not os.path.isfile(join(cell_folder, 'mods', os.path.split(nmodl)[-1])):
                if verbose:
                    print('cp {} {}'.format(nmodl, join(cell_folder, 'mods')))
                os.system('cp {} {}'.format(nmodl, join(cell_folder, 'mods')))
    starting_dir = os.getcwd()
    os.chdir(join(cell_folder, 'mods'))
    os.system('nrnivmodl')
    os.chdir(starting_dir)


def return_cell(cell_folder, model_type, cell_name, end_T, dt, start_T, verbose=False):
    """ Function to load cell models
    
    Parameters:
    -----------
    cell_folder : string
        Path to folder where cell model is saved.
    model_type : string
        Cell model type (e.g. 'bbp' for Human Brain Project)    
    cell_name : string
        Name of the cell
    end_T : float
        Simulation length [ms]
    dt: float
        Time step of simulation [ms]
    start_T: float
        Simulation start time (recording starts at 0 ms)

    Returns:
    --------
    cell : object
        LFPy cell object
    """
    import mpi4py.MPI
    import LFPy
    import neuron
    neuron.h.load_file("stdrun.hoc")
    neuron.h.load_file("import3d.hoc")

    cwd = os.getcwd()
    os.chdir(cell_folder)
    if verbose:
        print("Simulating ", cell_name)

    if model_type == 'bbp':
        neuron.load_mechanisms('../mods')

        f = open("template.hoc", 'r')
        templatename = get_templatename(f)
        f.close()

        f = open("biophysics.hoc", 'r')
        biophysics = get_templatename(f)
        f.close()

        f = open("morphology.hoc", 'r')
        morphology = get_templatename(f)
        f.close()

        # get synapses template name
        f = open(join("synapses", "synapses.hoc"), 'r')
        synapses = get_templatename(f)
        f.close()

        neuron.h.load_file('constants.hoc')
        if not hasattr(neuron.h, morphology):
            neuron.h.load_file(1, "morphology.hoc")

        if not hasattr(neuron.h, biophysics):
            neuron.h.load_file(1, "biophysics.hoc")

        if not hasattr(neuron.h, synapses):
            # load synapses
            neuron.h.load_file(1, join('synapses', 'synapses.hoc'))

        if not hasattr(neuron.h, templatename):
            neuron.h.load_file(1, "template.hoc")

        morphologyfile = os.listdir('morphology')[0]  # glob('morphology\\*')[0]

        # Instantiate the cell(s) using LFPy
        cell = LFPy.TemplateCell(morphology=join('morphology', morphologyfile),
                                 templatefile=join('template.hoc'),
                                 templatename=templatename,
                                 templateargs=0,
                                 tstop=end_T,
                                 tstart=start_T,
                                 dt=dt,
                                 v_init=-70,
                                 pt3d=True,
                                 delete_sections=True,
                                 verbose=True)

    else:
        raise NotImplementedError('Cell model %s is not implemented' \
                                  % model_type)

    os.chdir(cwd)
    return cell


def find_spike_idxs(v, thresh=-30, find_max=30):
    """ Find spike indices
    
    Parameters:
    -----------
    v: array_like
        Membrane potential
    thresh: float (optional, default = -30)
        Threshold for spike detections
    find_max: int
        Number of sample to find spike maximum after detection crossing

    Returns:
    --------
    spikes : array_like
        Indices of spike peaks in the positive direction, i.e. spikes
    """
    spikes_th = [idx for idx in range(len(v) - 1) if v[idx] < thresh < v[idx + 1]]
    spikes = []
    for sp in spikes_th:
        max_idx = np.argmax(v[sp:sp + find_max])
        spikes.append(sp + max_idx)
    return spikes


def set_input(weight, dt, T, cell, delay, stim_length):
    """ Set current input synapse in soma
    
    Parameters:
    -----------
    weight : float
        Strength of input current [nA]
    dt : float
        Time step of simulation [ms]
    T : float
        Total simulation time [ms]
    cell : object
        Cell object from LFPy
    delay : float
        Delay for input,i.e. when to start the input [ms]
    stim_length: float
        Duration of injected current [ms]
    
    Returns:
    --------
    noiseVec :  NEURON vector
        NEURON vector of input current
    cell : object
        LFPy cell object
    synapse : NEURON synapse
        NEURON synapse object
    """

    tot_ntsteps = int(round(T / dt + 1))

    I = np.ones(tot_ntsteps) * weight
    # I[stim_idxs] = weight
    noiseVec = neuron.h.Vector(I)
    syn = None
    for sec in cell.allseclist:
        if 'soma' in sec.name():
            # syn = neuron.h.ISyn(0.5, sec=sec) 
            syn = neuron.h.IClamp(0.5, sec=sec)
    syn.dur = stim_length
    syn.delay = delay  # cell.tstartms
    noiseVec.play(syn._ref_amp, dt)

    return noiseVec, cell, syn


def run_cell_model(cell_model, sim_folder, seed, verbose, **kwargs):
    """ Run simulation and adjust input strength to have a certain number of 
        spikes (target_spikes[0] < num_spikes <= target_spikes[1]
        where target_spikes=[10,30] by default)

    Parameters:
    -----------
    cell_model : string
        Path to folder where cell model is saved.
    model_type : string
        Cell model type (e.g. 'bbp')    
    sim_folder : string
        Data directory for transmembrane currents and membrane potential
        of the neuron.
    cell_model_id: int
        Arbitrary cell id, used to set the numpy.random seed.

    Returns:
    --------
    cell : object
        LFPy cell object
    """

    cell_name = os.path.split(cell_model)[-1]

    if not os.path.isdir(sim_folder):
        os.makedirs(sim_folder)

    imem_files = [f for f in os.listdir(sim_folder) if 'imem' in f]
    vmem_files = [f for f in os.listdir(sim_folder) if 'vmem' in f]

    if not (np.any([cell_name in ifile for ifile in imem_files]) and
            np.any([cell_name in vfile for vfile in vmem_files])):

        np.random.seed(seed)
        T = kwargs['sim_time'] * 1000
        dt = kwargs['dt']
        cell = return_cell(cell_model, 'bbp', cell_name, T, dt, 0)

        delay = kwargs['delay']
        stim_length = T - delay
        weights = kwargs['weights']
        weight = weights[0]
        target_spikes = kwargs['target_spikes']
        cuts = kwargs['cut_out']
        cut_out = [cuts[0] / dt, cuts[1] / dt]

        num_spikes = 0

        i = 0
        while not target_spikes[0] < num_spikes <= target_spikes[1]:
            noiseVec, cell, syn = set_input(weight, dt, T, cell, delay, stim_length)
            cell.simulate(rec_imem=True)

            t = cell.tvec
            v = cell.somav
            t = t
            v = v

            spikes = find_spike_idxs(v[int(cut_out[0]):-int(cut_out[1])])
            spikes = list(np.array(spikes) + cut_out[0])
            num_spikes = len(spikes)

            if verbose:
                print("Input weight: ", weight, " - Num Spikes: ", num_spikes)
            if num_spikes >= target_spikes[1]:
                weight *= weights[0]
            elif num_spikes <= target_spikes[0]:
                weight *= weights[1]

            i += 1
            if i >= 10:
                sys.exit()

        t = t[0:(int(cut_out[0]) + int(cut_out[1]))] - t[int(cut_out[0])]
        # discard first spike
        i_spikes = np.zeros((num_spikes - 1, cell.totnsegs, len(t)))
        v_spikes = np.zeros((num_spikes - 1, len(t)))

        for idx, spike_idx in enumerate(spikes[1:]):
            spike_idx = int(spike_idx)
            v_spike = v[spike_idx - int(cut_out[0]):spike_idx + int(cut_out[1])]
            i_spike = cell.imem[:, spike_idx - int(cut_out[0]):spike_idx + int(cut_out[1])]
            i_spikes[idx, :, :] = i_spike
            v_spikes[idx, :] = v_spike

        if not os.path.isdir(sim_folder):
            os.makedirs(sim_folder)
        np.save(join(sim_folder, 'imem_%d_%s.npy' % (num_spikes - 1, cell_name)), i_spikes)
        np.save(join(sim_folder, 'vmem_%d_%s.npy' % (num_spikes - 1, cell_name)), v_spikes)

    else:
        if verbose:
            print('\n\n\nCell has already be simulated. Using stored membrane currents\n\n\n')


def calc_extracellular(cell_model, save_sim_folder, load_sim_folder, seed, verbose=False, position=None, **kwargs):
    """  Loads data from previous cell simulation, and use results to generate
         arbitrary number of spikes above a certain noise level.

    Parameters:
    -----------
    cell_model : string
        Path to folder where cell model is saved.
    model_type : string
        Cell model type (e.g. 'bbp')    
    save_sim_folder : string
        Path to folder where to save EAP data
    load_sim_folder : string
        Path to folder from which  NEURON simulation results (currents, 
        membrane potential) are loaded
    rotation: string
        Type of rotation to apply to neuron morphologies 
        ('Norot','physrot','3drot')
    cell_model_id: int
        Arbitrary cell id, used to set the numpy.random seed.

    Returns:
    --------
        nothing, but saves the result
    """
    cell_name = os.path.split(cell_model)[-1]
    cell_save_name = cell_name
    np.random.seed(seed)

    T = kwargs['sim_time'] * 1000
    dt = kwargs['dt']
    rotation = kwargs['rot']
    nobs = kwargs['n']
    ncontacts = kwargs['ncontacts']
    overhang = kwargs['overhang']
    offset = kwargs['offset']
    x_lim = kwargs['xlim']
    y_lim = kwargs['ylim']
    z_lim = kwargs['zlim']
    min_amp = kwargs['min_amp']
    MEAname = kwargs['probe']
    offset = kwargs['offset']
    drifting = kwargs['drifting']
    if drifting:
        max_drift = kwargs['max_drift']
        min_drift = kwargs['min_drift']
        drift_steps = kwargs['drift_steps']
        drift_x_lim = kwargs['drift_xlim']
        drift_y_lim = kwargs['drift_ylim']
        drift_z_lim = kwargs['drift_zlim']

    sim_folder = join(save_sim_folder, rotation)
    cell = return_cell(cell_model, 'bbp', cell_name, T, dt, 0)

    # Load data from previous cell simulation
    imem_file = [f for f in os.listdir(load_sim_folder) if cell_name in f and 'imem' in f][0]
    vmem_file = [f for f in os.listdir(load_sim_folder) if cell_name in f and 'vmem' in f][0]
    i_spikes = np.load(join(load_sim_folder, imem_file))
    v_spikes = np.load(join(load_sim_folder, vmem_file))
    cell.tvec = np.arange(i_spikes.shape[-1]) * dt

    save_spikes = []
    save_pos = []
    save_rot = []
    save_offs = []
    target_num_spikes = int(nobs)

    # load MEA info
    elinfo = mu.return_mea_info(electrode_name=MEAname)

    # Create save folder
    save_folder = join(sim_folder, 'tmp_%d_%s' % (target_num_spikes, MEAname))

    if not os.path.isdir(save_folder):
        os.makedirs(save_folder)

    if verbose:
        print('Cell ', cell_save_name, ' extracellular spikes to be simulated')

    mea = mu.return_mea(info=elinfo)
    pos = mea.positions

    elec_x = pos[:, 0]
    elec_y = pos[:, 1]
    elec_z = pos[:, 2]

    N = np.empty((pos.shape[0], 3))
    for i in np.arange(N.shape[0]):
        N[i,] = [1, 0, 0]  # normal vec. of contacts

    # Add square electrodes (instead of circles)
    if ncontacts > 1:
        electrode_parameters = {
            'sigma': 0.3,  # extracellular conductivity
            'x': elec_x,  # x,y,z-coordinates of contact points
            'y': elec_y,
            'z': elec_z,
            'n': ncontacts,
            'r': elinfo['size'],
            'N': N,
            'contact_shape': elinfo['shape']
        }
    else:
        electrode_parameters = {
            'sigma': 0.3,  # extracellular conductivity
            'x': elec_x,  # x,y,z-coordinates of contact points
            'y': elec_y,
            'z': elec_z
        }

    if x_lim is None:
        x_lim = [float(np.min(elec_x) - overhang),
                 float(np.max(elec_x) + overhang)]
    if y_lim is None:
        y_lim = [float(np.min(elec_y) - overhang),
                 float(np.max(elec_y) + overhang)]
    if z_lim is None:
        z_lim = [float(np.min(elec_z) - overhang),
                 float(np.max(elec_z) + overhang)]

    ignored = 0
    saved = 0
    i = 0

    while len(save_spikes) < target_num_spikes:
        if i > 1000 * target_num_spikes:
            if verbose:
                print("Gave up finding spikes above noise level for %s" % cell_name)
            break
        spike_idx = np.random.randint(0, i_spikes.shape[0])  # Each cell has several spikes to choose from
        cell.imem = i_spikes[spike_idx, :, :]
        cell.somav = v_spikes[spike_idx, :]

        if not drifting:
            espikes, pos, rot, offs = return_extracellular_spike(cell=cell, cell_name=cell_name, model_type='bbp',
                                                                 electrode_parameters=electrode_parameters,
                                                                 limits=[x_lim, y_lim, z_lim], rotation=rotation,
                                                                 pos=position)
            # Method of Images for semi-infinite planes
            if elinfo['type'] == 'mea':
                espikes = espikes * 2

            if check_espike(espikes, min_amp):
                save_spikes.append(espikes)
                save_pos.append(pos)
                save_rot.append(rot)
                save_offs.append(offs)
                plot_spike = False
                if verbose:
                    print('Cell: ' + cell_name + ' Progress: [' +
                          str(len(save_spikes)) + '/' + str(target_num_spikes) + ']')
                saved += 1
        else:
            espikes, pos, rot, offs = return_extracellular_spike(cell, cell_name, 'bbp', electrode_parameters,
                                                                 [x_lim, y_lim, z_lim], rotation, pos=position)

            # Method of Images for semi-infinite planes
            if elinfo['type'] == 'mea':
                espikes = espikes * 2

            if pos[0] - drift_x_lim[0] > x_lim[0] and pos[0] - drift_x_lim[1] < x_lim[1] and \
                    pos[1] - drift_y_lim[0] > y_lim[0] and pos[1] - drift_y_lim[1] < y_lim[1] and \
                    pos[2] - drift_z_lim[0] > z_lim[0] and pos[2] - drift_z_lim[1] < z_lim[1]:
                if check_espike(espikes, min_amp):
                    drift_ok = False
                    # fix rotation while drifting
                    cell.set_rotation(rot[0], rot[1], rot[2])
                    max_trials = 100
                    tr = 0
                    while not drift_ok and tr < max_trials:
                        init_pos = pos
                        # find drifting final position within drift limits
                        x_rand = np.random.uniform(init_pos[0] + drift_x_lim[0], init_pos[0] + drift_x_lim[1])
                        y_rand = np.random.uniform(init_pos[1] + drift_y_lim[0], init_pos[1] + drift_y_lim[1])
                        z_rand = np.random.uniform(init_pos[2] + drift_z_lim[0], init_pos[2] + drift_z_lim[1])
                        final_pos = [x_rand, y_rand, z_rand]
                        drift_dist = np.linalg.norm(np.array(init_pos) - np.array(final_pos))

                        # check location and boundaries
                        if drift_dist < max_drift and drift_dist > min_drift and \
                                x_rand > x_lim[0] and x_rand < x_lim[1] and \
                                y_rand > y_lim[0] and y_rand < y_lim[1] and \
                                z_rand > z_lim[0] and z_rand < z_lim[1]:
                            espikes, pos, rot_, offs = return_extracellular_spike(cell, cell_name, 'bbp',
                                                                                  electrode_parameters,
                                                                                  [x_lim, y_lim, z_lim],
                                                                                  rotation=None,
                                                                                  pos=final_pos)
                            # check final position spike amplitude
                            if check_espike(espikes, min_amp):
                                if verbose:
                                    print('Found final drifting position')
                                drift_ok = True
                            else:
                                tr += 1
                                pass
                        else:
                            tr += 1
                            pass

                    # now compute drifting templates
                    if drift_ok:
                        drift_spikes = []
                        drift_pos = []
                        drift_rot = []
                        drift_dist = np.linalg.norm(np.array(init_pos) - np.array(final_pos))
                        drift_dir = np.array(final_pos) - np.array(init_pos)
                        for i, dp in enumerate(np.linspace(0, 1, drift_steps)):
                            pos_drift = init_pos + dp * drift_dir
                            espikes, pos, r_, offs = return_extracellular_spike(cell, cell_name, 'bbp',
                                                                                electrode_parameters,
                                                                                [x_lim, y_lim, z_lim],
                                                                                rotation=None,
                                                                                pos=pos_drift)
                            drift_spikes.append(espikes)
                            drift_pos.append(pos)
                            drift_rot.append(rot)

                        # reverse rotation
                        rev_rot = [-r for r in rot]
                        cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order='zyx')

                        drift_spikes = np.array(drift_spikes)
                        drift_pos = np.array(drift_pos)
                        if verbose:
                            print('Drift done from ', init_pos, ' to ', final_pos, ' with ', drift_steps, ' steps')

                        save_spikes.append(drift_spikes)
                        save_pos.append(drift_pos)
                        save_rot.append(rot)
                        save_offs.append(offs)
                        if verbose:
                            print('Cell: ' + cell_name + ' Progress: [' + str(len(save_spikes)) + '/' +
                                  str(target_num_spikes) + ']')
                        saved += 1
                    else:
                        if verbose:
                            print('Discarded for trials')
                else:
                    pass
            else:
                if verbose:
                    print('Discarded position: ', pos)
        i += 1

    save_spikes = np.array(save_spikes)
    save_pos = np.array(save_pos)
    save_rot = np.array(save_rot)

    np.save(join(save_folder, 'eap-%s' % cell_save_name), save_spikes)
    np.save(join(save_folder, 'pos-%s' % cell_save_name), save_pos)
    np.save(join(save_folder, 'rot-%s' % cell_save_name), save_rot)


def check_espike(espikes, min_amp):
    '''
    Check extracellular spike amplitude and shape (neg peak > pos peak)

    Parameters
    ----------
    espike: np.array
        EAP (n_elec, n_samples)
    min_amp: float
        Minimum amplitude

    Returns
    -------
    valid: bool
        If True EAP is valid

    '''
    valid = True
    if np.max(np.abs(np.min(espikes))) < min_amp:
        valid = False
    if np.abs(np.min(espikes)) < np.abs(np.max(espikes)):
        valid = False
    return valid


def get_physrot_specs(cell_name, model):
    """  Return physrot specifications for cell types
    
    Parameters:
    -----------
    cell_name : string
        The name of the cell.

    Returns:
    --------
    polarlim : array_like
        lower and upper bound for the polar angle
    pref_orient : array_like
        3-dim vetor of preferred orientation 
    """
    if model == 'bbp':
        polarlim = {'BP': [0., 15.],
                    'BTC': None,  # [0.,15.],
                    'ChC': None,  # [0.,15.],
                    'DBC': None,  # [0.,15.],
                    'LBC': None,  # [0.,15.],
                    'MC': [0., 15.],
                    'NBC': None,
                    'NGC': None,
                    'SBC': None,
                    'STPC': [0., 15.],
                    'TTPC1': [0., 15.],
                    'TTPC2': [0., 15.],
                    'UTPC': [0., 15.]}
        # how it's implemented, the NMC y axis points into the pref_orient direction after rotation
        pref_orient = {'BP': [0., 0., 1.],
                       'BTC': None,  # [0.,0.,1.],
                       'ChC': None,  # [0.,0.,1.],
                       'DBC': None,  # [0.,0.,1.],
                       'LBC': None,  # [0.,0.,1.],
                       'MC': [0., 0., 1.],
                       'NBC': None,
                       'NGC': None,
                       'SBC': None,
                       'STPC': [0., 0., 1.],
                       'TTPC1': [0., 0., 1.],
                       'TTPC2': [0., 0., 1.],
                       'UTPC': [0., 0., 1.]}
        return polarlim[cell_name.split('_')[1]], pref_orient[cell_name.split('_')[1]]
    else:
        raise NotImplementedError('Cell model %s is not implemented' \
                                  % model)


def return_extracellular_spike(cell, cell_name, model_type,
                               electrode_parameters, limits, rotation, pos=None):
    """    Calculate extracellular spike on MEA 
           at random position relative to cell

    Parameters:
    -----------
    cell: object
        cell object from LFPy
    cell_name: string
        name of cell model
    electrode_parameters: dict
        parameters to initialize LFPy.RecExtElectrode
    limits: array_like
        boundaries for neuron locations, shape=(3,2)
    rotation: string 
        random rotation to apply to the neuron ('Norot', '3drot', 'physrot')
    pos: array_like, (optional, default None)
        Can be used to set the cell soma to a specific position. If ``None``,
        the random position is used.
    Returns:
    --------
    Extracellular spike for each MEA contact site
    """
    import LFPy

    def get_xyz_angles(R):
        ''' Get rotation angles for each axis from rotation matrix
        
        Parameters;
        -----------
        R : matrix
            3x3 rotation matrix

        Returns:
        --------
        R_z : float
        R_y : float
        R_x : float
            Three angles for rotations around axis, defined by R = R_z.R_y.R_x
        '''
        rot_x = np.arctan2(R[2, 1], R[2, 2])
        rot_y = np.arcsin(-R[2, 0])
        rot_z = np.arctan2(R[1, 0], R[0, 0])
        return rot_x, rot_y, rot_z

    def get_rnd_rot_Arvo():
        """ Generate uniformly distributed random rotation matrices
        see: 'Fast Random Rotation Matrices' by Arvo (1992)
        
        Returns:
        --------
        R : 3x3 matrix
            random rotation matrix
        """
        gamma = np.random.uniform(0, 2. * np.pi)
        rotation_z = np.matrix([[np.cos(gamma), -np.sin(gamma), 0],
                                [np.sin(gamma), np.cos(gamma), 0],
                                [0, 0, 1]])
        x = np.random.uniform(size=2)
        v = np.array([np.cos(2. * np.pi * x[0]) * np.sqrt(x[1]),
                      np.sin(2. * np.pi * x[0]) * np.sqrt(x[1]),
                      np.sqrt(1 - x[1])])
        H = np.identity(3) - 2. * np.outer(v, v)
        M = -np.dot(H, rotation_z)
        return M

    def check_solidangle(matrix, pre, post, polarlim):
        """ Check whether a matrix rotates the vector 'pre' into a region
            defined by 'polarlim' around the vector 'post'

        Parameters:
        -----------
        matrix : matrix
            3x3 rotation matrix
        pre : array_like
            3-dim vector to be rotated
        post : array_like
            axis of the cones defining the post-rotation region
        polarlim : [float,float]
            Angles specifying the opening of the inner and outer cone 
            (aperture = 2*polarlim),
            i.e. the angle between rotated pre vector and post vector has to ly 
            within these polar limits.  

        Returns:
        --------
        test : bool
            True if the vector np.dot(matrix,pre) lies inside the specified region.
        """
        postest = np.dot(matrix, pre)
        c = np.dot(post / np.linalg.norm(post), postest / np.linalg.norm(postest))
        if np.cos(np.deg2rad(polarlim[1])) <= c <= np.cos(np.deg2rad(polarlim[0])):
            return True
        else:
            return False

    electrodes = LFPy.RecExtElectrode(cell, **electrode_parameters)

    '''Rotate neuron'''
    if rotation == 'norot':
        if model_type == 'bbp':
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.
            y_rot_offset = 0
            z_rot_offset = 0
        x_rot = x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot_offset
    elif rotation == 'xrot':
        if model_type == 'bbp':
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.
            y_rot_offset = 0
            z_rot_offset = 0
        x_rot, _, _ = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot + x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot_offset
    elif rotation == 'yrot':
        if model_type == 'bbp':
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.
            y_rot_offset = 0
            z_rot_offset = 0
        _, y_rot, _ = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot_offset
        y_rot = y_rot + y_rot_offset
        z_rot = z_rot_offset
    elif rotation == 'zrot':
        if model_type == 'bbp':
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.
            y_rot_offset = 0
            z_rot_offset = 0
        _, _, z_rot = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot + z_rot_offset
    elif rotation == '3drot':
        if model_type == 'bbp':
            x_rot_offset = np.pi / 2.  # align neuron with z axis
            y_rot_offset = 0  # align neuron with z axis
            z_rot_offset = 0  # align neuron with z axis
        x_rot, y_rot, z_rot = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot + x_rot_offset
        y_rot = y_rot + y_rot_offset
        z_rot = z_rot + z_rot_offset
    elif rotation == 'physrot':
        polarlim, pref_orient = get_physrot_specs(cell_name, model_type)
        if model_type == 'bbp':
            x_rot_offset = np.pi / 2.  # align neuron with z axis
            y_rot_offset = 0  # align neuron with z axis
            z_rot_offset = 0  # align neuron with z axis
        while True:
            R = np.array(get_rnd_rot_Arvo())
            if polarlim is None or pref_orient is None:
                valid = True
            else:
                valid = check_solidangle(R, [0., 0., 1.], pref_orient, polarlim)
            if valid:
                x_rot, y_rot, z_rot = get_xyz_angles(R)
                x_rot = x_rot + x_rot_offset
                y_rot = y_rot + y_rot_offset
                z_rot = z_rot + z_rot_offset
                break
    else:
        rotation = None
        x_rot = 0
        y_rot = 0
        z_rot = 0

    if pos is None:
        '''Move neuron randomly'''
        x_rand = np.random.uniform(limits[0][0], limits[0][1])
        y_rand = np.random.uniform(limits[1][0], limits[1][1])
        z_rand = np.random.uniform(limits[2][0], limits[2][1])
        cell.set_pos(x_rand, y_rand, z_rand)
        pos = [x_rand, y_rand, z_rand]
    else:
        cell.set_pos(pos[0], pos[1], pos[2])
    cell.set_rotation(x=x_rot, y=y_rot, z=z_rot)
    rot = [x_rot, y_rot, z_rot]

    electrodes.calc_lfp()

    # Reverse rotation to bring cell back into initial rotation state
    if rotation is not None:
        rev_rot = [-r for r in rot]
        cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order='zyx')

    return 1000 * electrodes.LFP, pos, rot, electrodes.offsets


def str2bool(v):
    """ Transform string to bool
    
    Parameters:
    -----------
    v : str
    
    Returns:
    --------
    transformed_v, bool
        If v is any of ("yes", "true", "t", "1") (case insensitive) 
        ``True`` is returned, else ``False``
    """
    return v.lower() in ("yes", "true", "t", "1")


if __name__ == '__main__':

    if len(sys.argv) == 3 and sys.argv[1] == 'compile':
        cell_folder = sys.argv[2]
        compile_all_mechanisms(cell_folder)
        sys.exit(0)
    elif len(sys.argv) == 5:
        cell_model = sys.argv[1]
        intraonly = str2bool(sys.argv[2])
        params_path = sys.argv[3]
        verbose = str2bool(sys.argv[4])

        with open(params_path, 'r') as f:
            if use_loader:
                params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                params = yaml.load(f)

        sim_folder = params['templates_folder']
        cell_folder = params['cell_models_folder']
        rot = params['rot']

        extra_sim_folder = params['templates_folder']
        vm_im_sim_folder = join(params['templates_folder'], 'intracellular')

        print('Intracellular simulation: ', cell_model)
        run_cell_model(cell_model, vm_im_sim_folder, verbose=verbose, **params)
        if not intraonly:
            print('Extracellular simulation: ', cell_model)
            calc_extracellular(cell_model, extra_sim_folder, vm_im_sim_folder, verbose=verbose, **params)



import pytest
import os
import numpy as np
import unittest
import MEArec as mr
import tempfile
import shutil
import yaml
import os
import elephant.statistics as stat
import quantities as pq
from distutils.version import StrictVersion
import tempfile
from click.testing import CliRunner

from MEArec.cli import cli, default_config, set_cell_models_folder, set_recordings_folder, set_recordings_params, \
    set_templates_folder, set_templates_params, gen_templates, gen_recordings

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False


class TestGenerators(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        info, info_folder = mr.get_default_config()
        cell_models_folder = info['cell_models_folder']
        self.num_cells = len([f for f in os.listdir(cell_models_folder) if 'mods' not in f])
        self.n = 5
        with open(info['templates_params']) as f:
            if use_loader:
                templates_params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                templates_params = yaml.load(f)

        with open(info['recordings_params']) as f:
            if use_loader:
                rec_params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                rec_params = yaml.load(f)

        self.test_dir = tempfile.mkdtemp()

        templates_params['n'] = self.n
        templates_params['probe'] = 'Neuronexus-32'
        templates_folder = info['templates_folder']
        print('Generating non-drifting templates')
        self.tempgen = mr.gen_templates(cell_models_folder, templates_folder=templates_folder,
                                        params=templates_params, parallel=True, delete_tmp=True, verbose=True)
        self.templates_params = templates_params
        self.num_templates, self.num_chan, self.num_samples = self.tempgen.templates.shape

        templates_params['drifting'] = True
        templates_params['drift_steps'] = 5
        templates_params['rot'] = 'norot'
        print('Generating drifting templates')
        self.tempgen_drift = mr.gen_templates(cell_models_folder, templates_folder=templates_folder,
                                              params=templates_params, parallel=True, delete_tmp=True, verbose=True)
        self.templates_params_drift = templates_params
        self.num_steps_drift = self.tempgen_drift.templates.shape[1]

        print('Making test recordings to test load functions')
        mr.save_template_generator(self.tempgen, self.test_dir + '/templates.h5')
        mr.save_template_generator(self.tempgen_drift, self.test_dir + '/templates_drift.h5')

        ne = 2
        ni = 1
        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['recordings']['modulation'] = 'none'
        rec_params['recordings']['filter'] = False
        rec_params['recordings']['extract_waveforms'] = True
        rec_params['recordings']['overlap'] = True
        rec_params['spiketrains']['duration'] = 2
        rec_params['templates']['min_dist'] = 1

        self.recgen = mr.gen_recordings(params=rec_params, tempgen=self.tempgen)
        self.recgen.annotate_overlapping_spike(parallel=False)
        mr.save_recording_generator(self.recgen, self.test_dir + '/recordings.h5')

    def test_gen_templates(self):
        print('Test templates generation')
        n = self.n
        num_cells = self.num_cells
        templates_params = self.templates_params

        assert self.tempgen.templates.shape[0] == (n * num_cells)
        assert len(self.tempgen.locations) == (n * num_cells)
        assert len(self.tempgen.rotations) == (n * num_cells)
        assert len(self.tempgen.celltypes) == (n * num_cells)
        assert len(np.unique(self.tempgen.celltypes)) == num_cells
        assert np.max(self.tempgen.locations[:, 0]) > templates_params['xlim'][0] \
               and np.max(self.tempgen.locations[:, 0]) < templates_params['xlim'][1]

    def test_gen_templates_drift(self):
        print('Test drifting templates generation')
        n = self.n
        num_cells = self.num_cells
        n_steps = self.num_steps_drift
        templates_params = self.templates_params_drift

        assert self.tempgen_drift.templates.shape[0] == (n * num_cells)
        assert self.tempgen_drift.locations.shape == (n * num_cells, n_steps, 3)
        assert len(self.tempgen_drift.rotations) == (n * num_cells)
        assert len(self.tempgen_drift.celltypes) == (n * num_cells)
        assert len(np.unique(self.tempgen_drift.celltypes)) == num_cells
        assert np.max(self.tempgen_drift.locations[:, :, 0]) > templates_params['xlim'][0] \
               and np.max(self.tempgen_drift.locations[:, :, 0]) < templates_params['xlim'][1]
        assert self.tempgen_drift.templates.shape[1] == self.num_steps_drift

    def test_gen_spiketrains(self):
        print('Test spike train generation')
        info, info_folder = mr.get_default_config()
        with open(info['recordings_params']) as f:
            if use_loader:
                rec_params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                rec_params = yaml.load(f)
        sp_params = rec_params['spiketrains']
        spgen = mr.SpikeTrainGenerator(sp_params)
        spgen.generate_spikes()

        #check ref period
        for st in spgen.all_spiketrains:
            isi = stat.isi(st).rescale('ms')
            assert np.all(isi.magnitude > sp_params['ref_per'])
            assert (1 / np.mean(isi.rescale('s'))) > sp_params['min_rate']

        sp_params['process'] = 'gamma'
        spgen = mr.SpikeTrainGenerator(sp_params)
        spgen.generate_spikes()
        for st in spgen.all_spiketrains:
            isi = stat.isi(st).rescale('ms')
            assert np.all(isi.magnitude > sp_params['ref_per'])
            assert (1 / np.mean(isi.rescale('s'))) > sp_params['min_rate']

    def test_gen_recordings_modulations(self):
        print('Test recording generation - modulation')
        info, info_folder = mr.get_default_config()
        ne = 1
        ni = 2
        fe = 30
        fi = 50
        num_chan = self.num_chan
        n_neurons = ne + ni

        with open(info['recordings_params']) as f:
            if use_loader:
                rec_params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                rec_params = yaml.load(f)

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['f_exc'] = fe
        rec_params['spiketrains']['f_inh'] = fi
        rec_params['spiketrains']['duration'] = 3
        n_jitter = [1, 5]
        modulations = ['none', 'electrode', 'template', 'template-isi', 'electrode-isi']
        rec_params['templates']['min_dist'] = 1
        bursting = [False, True]

        for mod in modulations:
            for b in bursting:
                for j in n_jitter:
                    print('Modulation: modulation', mod, 'bursting', b, 'jitter', j)
                    rec_params['templates']['n_jitters'] =j
                    rec_params['recordings']['modulation'] = mod
                    rec_params['recordings']['bursting'] = b

                    recgen_mod = mr.gen_recordings(params=rec_params, tempgen=self.tempgen)

                    assert recgen_mod.recordings.shape[0] == num_chan
                    assert len(recgen_mod.spiketrains) == n_neurons
                    assert recgen_mod.channel_positions.shape == (num_chan, 3)
                    if j == 1:
                        assert recgen_mod.templates.shape[:2] == (n_neurons, num_chan)
                    else:
                        assert recgen_mod.templates.shape[:3] == (n_neurons, j, num_chan)
                    assert recgen_mod.voltage_peaks.shape == (n_neurons, num_chan)
                    assert len(recgen_mod.spike_traces) == n_neurons
                    del recgen_mod

    def test_gen_recordings_mod_bursting_sync(self):
        print('Test recording generation - bursting')
        info, info_folder = mr.get_default_config()
        ne = 10
        ni = 5
        fe = 30
        fi = 50
        num_chan = self.num_chan
        n_neurons = ne + ni

        with open(info['recordings_params']) as f:
            if use_loader:
                rec_params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                rec_params = yaml.load(f)

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['f_exc'] = fe
        rec_params['spiketrains']['f_inh'] = fi
        rec_params['spiketrains']['duration'] = 3
        n_jitter = 4
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['recordings']['modulation'] = 'electrode-isi'
        rec_params['recordings']['bursting'] = True
        rec_params['recordings']['sync_rate'] = 0.2
        rec_params['recordings']['overlap'] = False
        rec_params['recordings']['extract_waveforms'] = True
        rec_params['templates']['min_dist'] = 1
        rec_params['templates']['min_amp'] = 30

        recgen_burst = mr.gen_recordings(params=rec_params, tempgen=self.tempgen)

        assert recgen_burst.recordings.shape[0] == num_chan
        assert len(recgen_burst.spiketrains) == n_neurons
        assert recgen_burst.channel_positions.shape == (num_chan, 3)
        assert recgen_burst.templates.shape[:3] == (n_neurons, n_jitter, num_chan)
        assert recgen_burst.voltage_peaks.shape == (n_neurons, num_chan)
        assert len(recgen_burst.spike_traces) == n_neurons

    def test_gen_recordings_noise(self):
        print('Test recording generation - noise')
        info, info_folder = mr.get_default_config()
        rates = [30, 50]
        types = ['E', 'I']
        num_chan = self.num_chan
        n_neurons = len(rates)

        with open(info['recordings_params']) as f:
            if use_loader:
                rec_params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                rec_params = yaml.load(f)

        rec_params['spiketrains']['rates'] = rates
        rec_params['spiketrains']['types'] = types
        rec_params['spiketrains']['duration'] = 3
        n_jitter = 3
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['templates']['min_dist'] = 1
        rec_params['recordings']['modulation'] = 'none'
        noise_modes = ['uncorrelated', 'distance-correlated']
        chunk_noise = [0, 2]
        noise_color = [True, False]

        for mode in noise_modes:
            for ch in chunk_noise:
                for color in noise_color:
                    print('Noise: mode', mode, 'chunks', ch, 'color', color)
                    rec_params['recordings']['noise_mode'] = mode
                    rec_params['recordings']['chunk_noise_duration'] = ch
                    rec_params['recordings']['noise_color'] = color
                    recgen_noise = mr.gen_recordings(params=rec_params, tempgen=self.tempgen)

                    assert recgen_noise.recordings.shape[0] == num_chan
                    assert len(recgen_noise.spiketrains) == n_neurons
                    assert recgen_noise.channel_positions.shape == (num_chan, 3)
                    assert recgen_noise.templates.shape[:3] == (n_neurons, n_jitter, num_chan)
                    assert recgen_noise.voltage_peaks.shape == (n_neurons, num_chan)
                    assert len(recgen_noise.spike_traces) == n_neurons
                    del recgen_noise

    def test_gen_recordings_only_noise(self):
        print('Test recording generation - only noise')
        info, info_folder = mr.get_default_config()
        ne = 0
        ni = 0
        num_chan = self.num_chan
        n_neurons = ne + ni

        with open(info['recordings_params']) as f:
            if use_loader:
                rec_params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                rec_params = yaml.load(f)

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = 3
        rec_params['cell_types'] = None
        n_jitter = 3
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['templates']['min_dist'] = 1
        rec_params['recordings']['modulation'] = 'none'
        rec_params['recordings']['noise_mode'] = 'distance-correlated'
        rec_params['recordings']['noise_color'] = True
        recgen_noise = mr.gen_recordings(params=rec_params, tempgen=self.tempgen)

        assert recgen_noise.recordings.shape[0] == num_chan
        assert recgen_noise.channel_positions.shape == (num_chan, 3)
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spike_traces) == n_neurons

    def test_gen_recordings_drift(self):
        print('Test recording generation - drift')
        info, info_folder = mr.get_default_config()
        ne = 1
        ni = 1
        num_chan = self.num_chan
        n_steps = self.num_steps_drift
        n_neurons = ne + ni

        with open(info['recordings_params']) as f:
            if use_loader:
                rec_params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                rec_params = yaml.load(f)

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = 3
        n_jitter = [1, 3]
        rec_params['recordings']['modulation'] = 'none'
        rec_params['recordings']['drifting'] = True
        rec_params['recordings']['drift_velocity'] = 300
        rec_params['templates']['min_dist'] = 1

        modulations = ['none', 'template', 'electrode']
        bursting = [False, True]

        for i, mod in enumerate(modulations):
            for b in bursting:
                for j in n_jitter:
                    print('Drifting: modulation', mod, 'bursting', b, 'jitter', j)
                    rec_params['templates']['n_jitters'] = j
                    rec_params['recordings']['modulation'] = mod
                    rec_params['recordings']['bursting'] = b
                    if i == len(modulations) - 1:
                        rec_params['recordings']['fs'] = 30
                        rec_params['recordings']['n_drifting'] = 1
                    recgen_drift = mr.gen_recordings(params=rec_params, tempgen=self.tempgen_drift)
                    assert recgen_drift.recordings.shape[0] == num_chan
                    assert len(recgen_drift.spiketrains) == n_neurons
                    assert recgen_drift.channel_positions.shape == (num_chan, 3)
                    if j == 1:
                        assert recgen_drift.templates.shape[0] == n_neurons
                        assert recgen_drift.templates.shape[2] == num_chan
                    else:
                        assert recgen_drift.templates.shape[0] == n_neurons
                        assert recgen_drift.templates.shape[2] == j
                        assert recgen_drift.templates.shape[3] == num_chan
                    assert len(recgen_drift.spike_traces) == n_neurons
                    del recgen_drift

    def test_save_load_templates(self):
        tempgen = mr.load_templates(self.test_dir + '/templates.h5', verbose=True)
        tempgen_drift = mr.load_templates(self.test_dir + '/templates_drift.h5')
        tempgen_drift_f = mr.load_templates(self.test_dir + '/templates_drift.h5', return_h5_objects=True)

        assert np.allclose(tempgen.templates, self.tempgen.templates)
        assert np.allclose(tempgen.locations, self.tempgen.locations)
        assert np.allclose(tempgen.rotations, self.tempgen.rotations)
        assert np.allclose(tempgen_drift.templates, self.tempgen_drift.templates)
        assert np.allclose(tempgen_drift.locations, self.tempgen_drift.locations)
        assert np.allclose(tempgen_drift.rotations, self.tempgen_drift.rotations)
        assert np.allclose(tempgen_drift.templates, np.array(tempgen_drift_f.templates))

    def test_save_load_recordings(self):
        recgen_loaded = mr.load_recordings(self.test_dir + '/recordings.h5', verbose=True)
        recgen_loaded_f = mr.load_recordings(self.test_dir + '/recordings.h5', return_h5_objects=True, verbose=True)

        assert np.allclose(recgen_loaded.templates, self.recgen.templates)
        assert np.allclose(recgen_loaded.recordings, self.recgen.recordings)
        assert np.allclose(recgen_loaded.spike_traces, self.recgen.spike_traces)
        assert np.allclose(recgen_loaded.voltage_peaks, self.recgen.voltage_peaks)
        assert np.allclose(recgen_loaded.channel_positions, self.recgen.channel_positions)
        assert np.allclose(recgen_loaded.timestamps, self.recgen.timestamps.magnitude)
        assert np.allclose(recgen_loaded.recordings, np.array(recgen_loaded_f.recordings))

    def test_plots(self):
        _ = mr.plot_rasters(self.recgen.spiketrains)
        _ = mr.plot_rasters(self.recgen.spiketrains, overlap=True)
        _ = mr.plot_rasters(self.recgen.spiketrains, bintype=True)
        _ = mr.plot_rasters(self.recgen.spiketrains, color='g')
        _ = mr.plot_rasters(self.recgen.spiketrains, color=['g']*len(self.recgen.spiketrains))
        _ = mr.plot_recordings(self.recgen)
        _ = mr.plot_templates(self.recgen)
        _ = mr.plot_templates(self.recgen, single_axes=True)
        _ = mr.plot_waveforms(self.recgen)
        _ = mr.plot_waveforms(self.recgen, electrode=0)
        _ = mr.plot_waveforms(self.recgen, electrode='max')
        self.recgen.spiketrains[0].waveforms = None
        _ = mr.plot_waveforms(self.recgen, electrode='max', color_isi=True)

    def test_extract_features(self):
        feat_t0 = mr.get_templates_features(self.tempgen.templates, feat_list=['na', 'rep', 'amp', 'width', 'fwhm',
                                                                               'ratio', 'speed'],
                                            dt=self.tempgen.info['params']['dt'])
        assert feat_t0['na'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['rep'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['amp'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['width'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['fwhm'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['ratio'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['speed'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])

    def test_cli(self):
        default_config, mearec_home = mr.get_default_config()

        runner = CliRunner()
        result = runner.invoke(cli, ["--help"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["default-config"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["gen-templates", '-n', '2', '--no-parallel', '-r', '3drot', '-nc', '2',
                                     '-ov', '20', '-off', '0', '-dt', '10', '-s', '1', '-mind', '10', '-maxd', '100',
                                     '-drst', '10', '-v'])
        print(result.output)
        assert result.exit_code == 0
        result = runner.invoke(cli, ["gen-recordings", '-t', self.test_dir + '/templates.h5', '-ne', '2', '-ni', '1',
                                     '-fe', '5', '-fi', '15', '-se', '1', '-si', '1', '-mr', '0.2',
                                     '-rp', '2', '-p', 'poisson', '-md', '1', '-mina', '10', '-maxa', '1000',
                                     '--fs', '32',  '-sr', '0', '-sj', '1', '-nl', '10', '-m', 'none',
                                     '-chn', '0', '-chf', '0', '-nseed', '10', '-hd', '30', '-cn', '-cp', '500',
                                     '-cq', '1', '-rnf', '1', '-stseed', '100', '-tseed', '10',
                                     '--no-filt', '-fc', '500', '-fo', '3', '--overlap', '-ot', '0.8', '--extract-wf',
                                     '-angt', '15', '-drvel', '10', '-tsd', '1', '-v'])
        print(result.output)
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-params", '.'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-params", default_config['templates_params']])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-params", '.'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-params", default_config['recordings_params']])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-cell-models-folder", './cell_models'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-cell-models-folder", default_config['cell_models_folder']])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", '.'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", '.'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", './templates', '--create'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", './recordings', '--create'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", default_config['templates_folder']])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", default_config['recordings_folder']])
        assert result.exit_code == 0


# if __name__ == '__main__':
#     unittest.main()



from __future__ import print_function, division

import numpy as np
import neo
import elephant.spike_train_generation as stg
import elephant.conversion as conv
import elephant.statistics as stat
import scipy.signal as ss
import time
from copy import copy, deepcopy
from MEArec.tools import *
import MEAutility as mu
import threading
import shutil
import yaml
from pprint import pprint
import quantities as pq
from quantities import Quantity
from distutils.version import StrictVersion

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False


class simulationThread(threading.Thread):
    def __init__(self, threadID, name, simulate_script, numb, tot, cell_model,
                 model_folder, intraonly, params, verbose):
        threading.Thread.__init__(self)
        self.threadID = threadID
        self.name = name
        self.sim_script = simulate_script
        self.numb = numb
        self.tot = tot
        self.cell_model = cell_model
        self.model_folder = model_folder
        self.intra = intraonly
        self.params = params
        self.verbose = verbose

    def run(self):
        print("Starting " + self.name)
        print('\n\n', self.cell_model, self.numb + 1, '/', self.tot, '\n\n')
        os.system('python %s %s %s %s %s' \
                  % (self.sim_script, join(self.model_folder, self.cell_model), self.intra, self.params, self.verbose))
        print("Exiting " + self.name)


class TemplateGenerator:
    '''
    Class for generation of templates called by the gen_templates function.
    The list of parameters is in default_params/templates_params.yaml.
    '''
    def __init__(self, cell_models_folder=None, templates_folder=None, temp_dict=None, info=None,
                 params=None, intraonly=False, parallel=True, delete_tmp=True, verbose=False):
        self.verbose = verbose
        if temp_dict is not None and info is not None:
            if 'templates' in temp_dict.keys():
                self.templates = temp_dict['templates']
            if 'locations' in temp_dict.keys():
                self.locations = temp_dict['locations']
            if 'rotations' in temp_dict.keys():
                self.rotations = temp_dict['rotations']
            if 'celltypes' in temp_dict.keys():
                self.celltypes = temp_dict['celltypes']
            self.info = info
        else:
            if cell_models_folder is None:
                raise AttributeError("Specify cell folder!")
            if params is None:
                if self.verbose:
                    print("Using default parameters")
                self.params = {}
            else:
                self.params = deepcopy(params)
            self.cell_model_folder = cell_models_folder
            self.templates_folder = templates_folder
            self.simulation_params = {'intraonly': intraonly, 'parallel': parallel, 'delete_tmp': delete_tmp}

    def generate_templates(self):
        '''
        Generate templates.
        '''
        cell_models_folder = self.cell_model_folder
        templates_folder = self.templates_folder
        intraonly = self.simulation_params['intraonly']
        parallel = self.simulation_params['parallel']
        delete_tmp = self.simulation_params['delete_tmp']

        if os.path.isdir(cell_models_folder):
            cell_models = [f for f in os.listdir(join(cell_models_folder)) if 'mods' not in f]
            if len(cell_models) == 0:
                raise AttributeError(cell_models_folder, ' contains no cell models!')
        else:
            raise NotADirectoryError('Cell models folder: does not exist!')

        this_dir, this_filename = os.path.split(__file__)
        simulate_script = join(this_dir, 'simulate_cells.py')
        self.params['cell_models_folder'] = cell_models_folder
        self.params['templates_folder'] = templates_folder

        # Compile NEURON models (nrnivmodl)
        if not os.path.isdir(join(cell_models_folder, 'mods')):
            if self.verbose:
                print('Compiling NEURON models')
            os.system('python %s compile %s' % (simulate_script, cell_models_folder))

        if 'sim_time' not in self.params.keys():
            self.params['sim_time'] = 1
        if 'target_spikes' not in self.params.keys():
            self.params['target_spikes'] = [3, 50]
        if 'cut_out' not in self.params.keys():
            self.params['cut_out'] = [2, 5]
        if 'dt' not in self.params.keys():
            self.params['dt'] = 2 ** -5
        if 'delay' not in self.params.keys():
            self.params['delay'] = 10
        if 'weights' not in self.params.keys():
            self.params['weights'] = [0.25, 1.75]

        if 'rot' not in self.params.keys():
            self.params['rot'] = 'physrot'
        if 'probe' not in self.params.keys():
            available_mea = mu.return_mea_list()
            probe = available_mea[np.random.randint(len(available_mea))]
            if self.verbose:
                print("Probe randomly set to: %s" % probe)
            self.params['probe'] = probe
        if 'ncontacts' not in self.params.keys():
            self.params['ncontacts'] = 1
        if 'overhang' not in self.params.keys():
            self.params['overhang'] = 1
        if 'xlim' not in self.params.keys():
            self.params['xlim'] = [10, 80]
        if 'ylim' not in self.params.keys():
            self.params['ylim'] = None
        if 'zlim' not in self.params.keys():
            self.params['zlim'] = None
        if 'offset' not in self.params.keys():
            self.params['offset'] = 0
        if 'det_thresh' not in self.params.keys():
            self.params['det_thresh'] = 30
        if 'n' not in self.params.keys():
            self.params['n'] = 50
        if 'seed' not in self.params.keys():
            self.params['seed'] = np.random.randint(1, 10000)
        elif self.params['seed'] is None:
            self.params['seed'] = np.random.randint(1, 10000)
        if templates_folder is None:
            self.params['templates_folder'] = os.getcwd()
            templates_folder = self.params['templates_folder']
        else:
            self.params['templates_folder'] = templates_folder
        if 'drifting' not in self.params.keys():
            self.params['drifting'] = False

        rot = self.params['rot']
        n = self.params['n']
        probe = self.params['probe']

        tmp_params_path = 'tmp_params_path'
        with open(tmp_params_path, 'w') as f:
            yaml.dump(self.params, f)

        # Simulate neurons and EAP for different cell models sparately
        if parallel:
            start_time = time.time()
            if self.verbose:
                print('Parallel')
            tot = len(cell_models)
            threads = []
            for numb, cell_model in enumerate(cell_models):
                threads.append(simulationThread(numb, "Thread-" + str(numb), simulate_script,
                                                numb, tot, cell_model, cell_models_folder, intraonly,
                                                tmp_params_path, self.verbose))
            for t in threads:
                t.start()
            for t in threads:
                t.join()
            print('\n\n\nSimulation time: ', time.time() - start_time, '\n\n\n')
        else:
            start_time = time.time()
            for numb, cell_model in enumerate(cell_models):
                if self.verbose:
                    print('\n\n', cell_model, numb + 1, '/', len(cell_models), '\n\n')
                os.system('python %s %s %s %s %s' \
                          % (simulate_script, join(cell_models_folder, cell_model), intraonly, tmp_params_path,
                             self.verbose))
            print('\n\n\nSimulation time: ', time.time() - start_time, '\n\n\n')

        tmp_folder = join(templates_folder, rot, 'tmp_%d_%s' % (n, probe))
        templates, locations, rotations, celltypes = load_tmp_eap(tmp_folder)
        if delete_tmp:
            shutil.rmtree(tmp_folder)
            os.remove(tmp_params_path)

        self.info = {}

        self.templates = templates
        self.locations = locations
        self.rotations = rotations
        self.celltypes = celltypes
        self.info['params'] = self.params
        self.info['electrodes'] = mu.return_mea_info(probe)


class SpikeTrainGenerator:
    '''
    Class for generation of spike trains called by the gen_recordings function.
    The list of parameters is in default_params/recordings_params.yaml (spiketrains field).
    '''
    def __init__(self, params=None, spiketrains=None, verbose=False):
        self.verbose = verbose
        if params is None:
            if self.verbose:
                print("Using default parameters")
            self.params = {}
        if spiketrains is None:
            self.params = deepcopy(params)
            if self.verbose:
                print('Spiketrains seed: ', self.params['seed'])
            np.random.seed(self.params['seed'])

            if 't_start' not in self.params.keys():
                params['t_start'] = 0
            self.params['t_start'] = params['t_start'] * pq.s
            if 'duration' not in self.params.keys():
                params['duration'] = 10
            self.params['t_stop'] = self.params['t_start'] + params['duration'] * pq.s
            if 'min_rate' not in self.params.keys():
                params['min_rate'] = 0.1
            self.params['min_rate'] = params['min_rate'] * pq.Hz
            if 'ref_per' not in self.params.keys():
                params['ref_per'] = 2
            self.params['ref_per'] = params['ref_per'] * pq.ms
            if 'process' not in self.params.keys():
                params['process'] = 'poisson'
            self.params['process'] = params['process']
            if 'gamma_shape' not in self.params.keys() and params['process'] == 'gamma':
                params['gamma_shape'] = 2
                self.params['gamma_shape'] = params['gamma_shape']

            if 'rates' in self.params.keys():  # all firing rates are provided
                self.params['rates'] = self.params['rates'] * pq.Hz
                self.n_neurons = len(self.params['rates'])
            else:
                rates = []
                types = []
                if 'f_exc' not in self.params.keys():
                    params['f_exc'] = 5
                self.params['f_exc'] = params['f_exc'] * pq.Hz
                if 'f_inh' not in self.params.keys():
                    params['f_inh'] = 15
                self.params['f_inh'] = params['f_inh'] * pq.Hz
                if 'st_exc' not in self.params.keys():
                    params['st_exc'] = 1
                self.params['st_exc'] = params['st_exc'] * pq.Hz
                if 'st_inh' not in self.params.keys():
                    params['st_inh'] = 3
                self.params['st_inh'] = params['st_inh'] * pq.Hz
                if 'n_exc' not in self.params.keys():
                    params['n_exc'] = 15
                self.params['n_exc'] = params['n_exc']
                if 'n_inh' not in self.params.keys():
                    params['n_inh'] = 5
                self.params['n_inh'] = params['n_inh']

                for exc in np.arange(self.params['n_exc']):
                    rate = self.params['st_exc'] * np.random.randn() + self.params['f_exc']
                    if rate < self.params['min_rate']:
                        rate = self.params['min_rate']
                    rates.append(rate)
                    types.append('e')
                for inh in np.arange(self.params['n_inh']):
                    rate = self.params['st_inh'] * np.random.randn() + self.params['f_inh']
                    if rate < self.params['min_rate']:
                        rate = self.params['min_rate']
                    rates.append(rate)
                    types.append('i')
                self.params['rates'] = rates
                self.params['types'] = types
                self.n_neurons = len(self.params['rates'])

            self.changing = False
            self.intermittent = False

            self.info = params
            self.spiketrains = False
        else:
            self.all_spiketrains = spiketrains
            self.spiketrains = True

    def set_spiketrain(self, idx, spiketrain):
        '''
        Sets spike train idx to new spiketrain.

        Parameters
        ----------
        idx : int
            Index of spike train to set
        spiketrain : neo.SpikeTrain
            New spike train

        '''
        self.all_spiketrains[idx] = spiketrain

    def generate_spikes(self):
        '''
        Generate spike trains based on default_params of the SpikeTrainGenerator class.
        self.all_spiketrains contains the newly generated spike trains
        '''

        if not self.spiketrains:
            self.all_spiketrains = []
            idx = 0
            for n in np.arange(self.n_neurons):
                if not self.changing and not self.intermittent:
                    rate = self.params['rates'][n]
                    if self.params['process'] == 'poisson':
                        st = stg.homogeneous_poisson_process(rate,
                                                             self.params['t_start'], self.params['t_stop'])
                    elif self.params['process'] == 'gamma':
                        st = stg.homogeneous_gamma_process(self.params['gamma_shape'], rate,
                                                           self.params['t_start'], self.params['t_stop'])
                else:
                    raise NotImplementedError('Changing and intermittent spiketrains are not impleented yet')
                self.all_spiketrains.append(st)
                self.all_spiketrains[-1].annotate(fr=rate)
                if 'n_exc' in self.params.keys() and 'n_inh' in self.params.keys():
                    if idx < self.params['n_exc']:
                        self.all_spiketrains[-1].annotate(type='E')
                    else:
                        self.all_spiketrains[-1].annotate(type='I')
                idx += 1

            # check consistency and remove spikes below refractory period
            for idx, st in enumerate(self.all_spiketrains):
                isi = stat.isi(st)
                idx_remove = np.where(isi < self.params['ref_per'])[0]
                spikes_to_remove = len(idx_remove)
                unit = st.times.units

                while spikes_to_remove > 0:
                    new_times = np.delete(st.times, idx_remove[0]) * unit
                    st = neo.SpikeTrain(new_times, t_start=self.params['t_start'], t_stop=self.params['t_stop'])
                    isi = stat.isi(st)
                    idx_remove = np.where(isi < self.params['ref_per'])[0]
                    spikes_to_remove = len(idx_remove)

                st.annotations = self.all_spiketrains[idx].annotations
                self.set_spiketrain(idx, st)
        else:
            print("SpikeTrainGenerator initialized with existing spike trains!")

    def add_synchrony(self, idxs, rate=0.05, time_jitt=1*pq.ms, verbose=False):
        '''
        Adds synchronous spikes between pairs of spike trains at a certain rate.

        Parameters
        ----------
        idxs : list or array
            Spike train indexes to add synchrony to
        rate : float
            Rate of added synchrony spike to spike train idxs[1] for each spike of idxs[0]
        time_jitt : quantity
            Maximum time jittering between added spikes
        verbose : bool
            If True output is verbose

        Returns
        -------
        sync_rate : float
            New synchrony rate
        fr1 : quantity
            Firing rate spike train 1
        fr2 : quantity
            Firing rate spike train 2

        '''
        idx1 = idxs[0]
        idx2 = idxs[1]
        st1 = self.all_spiketrains[idx1]
        st2 = self.all_spiketrains[idx2]
        times1 = st1.times
        times2 = st2.times
        t_start = st2.t_start
        t_stop = st2.t_stop
        unit = times2.units
        added_spikes_t1 = 0
        added_spikes_t2 = 0

        all_times_shuffle = np.concatenate((times1, times2))
        all_times_shuffle = all_times_shuffle[np.random.permutation(len(all_times_shuffle))] * unit

        sync_rate = compute_sync_rate(st1, st2, time_jitt)
        for t in all_times_shuffle:
            if sync_rate <= rate:
                # check time difference
                if t in times1:
                    t_diff = np.abs(t.rescale(pq.ms).magnitude - times2.rescale(pq.ms).magnitude)
                    if np.all(t_diff > self.params['ref_per']):
                        t1_jitt = time_jitt.rescale(unit).magnitude * np.random.rand(1) + t.rescale(unit).magnitude - \
                                  (time_jitt.rescale(unit) / 2).magnitude
                        times2 = np.sort(np.concatenate((np.array(times2), np.array(t1_jitt))))
                        times2 = times2 * unit
                        st2 = neo.SpikeTrain(times2, t_start=t_start, t_stop=t_stop)
                        added_spikes_t1 += 1
                elif t in times2:
                    t_diff = np.abs(t.rescale(pq.ms).magnitude - times1.rescale(pq.ms).magnitude)
                    if np.all(t_diff > self.params['ref_per']):
                        t2_jitt = time_jitt.rescale(unit).magnitude * np.random.rand(1) + t.rescale(unit).magnitude - \
                                  (time_jitt.rescale(unit) / 2).magnitude
                        times1 = np.sort(np.concatenate((np.array(times1), np.array(t2_jitt))))
                        times1 = times1 * unit
                        st1 = neo.SpikeTrain(times1, t_start=t_start, t_stop=t_stop)
                        added_spikes_t2 += 1
                sync_rate = compute_sync_rate(st1, st2, time_jitt)
            else:
                break
        st1.annotations = self.all_spiketrains[idx1].annotations
        st2.annotations = self.all_spiketrains[idx2].annotations
        self.set_spiketrain(idx1, st1)
        self.set_spiketrain(idx2, st2)
        if verbose:
            print("Added", added_spikes_t1, "spikes to spike train", idxs[0],
                  "and", added_spikes_t2, "spikes to spike train", idxs[1])

        fr1 = len(st1.times) / st1.t_stop
        fr2 = len(st2.times) / st2.t_stop

        return sync_rate, fr1, fr2


class RecordingGenerator:
    '''
    Class for generation of recordings called by the gen_recordings function.
    The list of parameters is in default_params/recordings_params.yaml.
    '''
    def __init__(self, spgen=None, tempgen=None, params=None, rec_dict=None, info=None, verbose=True):
        self.verbose = verbose
        if rec_dict is not None and info is not None:
            if 'recordings' in rec_dict.keys():
                self.recordings = rec_dict['recordings']
            else:
                self.recordings = np.array([])
            if 'spiketrains' in rec_dict.keys():
                self.spiketrains = rec_dict['spiketrains']
            else:
                self.spiketrains = np.array([])
            if 'templates' in rec_dict.keys():
                self.templates = rec_dict['templates']
            else:
                self.templates = np.array([])
            if 'channel_positions' in rec_dict.keys():
                self.channel_positions = rec_dict['channel_positions']
            else:
                self.channel_positions = np.array([])
            if 'timestamps' in rec_dict.keys():
                self.timestamps = rec_dict['timestamps']
            else:
                self.timestamps = np.array([])
            if 'voltage_peaks' in rec_dict.keys():
                self.voltage_peaks = rec_dict['voltage_peaks']
            else:
                self.voltage_peaks = np.array([])
            if 'spike_traces' in rec_dict.keys():
                self.spike_traces = rec_dict['spike_traces']
            else:
                self.spike_traces = np.array([])
            self.info = info
        else:
            if spgen is None or tempgen is None:
                raise AttributeError("Specify SpikeGenerator and TemplateGenerator objects!")
            if params is None:
                if self.verbose:
                    print("Using default parameters")
                params = {'spiketrains': {},
                          'celltypes': {},
                          'templates': {},
                          'recordings': {}}
            self.params = deepcopy(params)
            self.spgen = spgen
            self.tempgen = tempgen

    def generate_recordings(self):
        '''
        Generate recordings.
        '''
        params = deepcopy(self.params)
        temp_params = self.params['templates']
        rec_params = self.params['recordings']
        st_params = self.params['spiketrains']
        celltype_params = self.params['cell_types']

        if 'rates' in st_params.keys():
            assert st_params['types'] is not None, "If 'rates' are provided as spiketrains parameters, " \
                                                   "corresponding 'types' ('E'-'I') must be provided"
            n_exc = st_params['types'].count('E')
            n_inh = st_params['types'].count('I')

        tempgen = self.tempgen
        spgen = self.spgen
        eaps = tempgen.templates
        locs = tempgen.locations
        rots = tempgen.rotations
        celltypes = tempgen.celltypes
        temp_info = tempgen.info

        spiketrains = spgen.all_spiketrains

        n_neurons = len(spiketrains)
        cut_outs = temp_info['params']['cut_out']
        if len(spiketrains) > 0:
            duration = spiketrains[0].t_stop - spiketrains[0].t_start
            only_noise = False
        else:
            if self.verbose:
                print('No spike trains provided: only simulating noise')
            only_noise = True
            duration = st_params['duration'] * pq.s

        if 'fs' not in rec_params.keys():
            params['recordings']['fs'] = 1. / temp_info['params']['dt']
            fs = params['recordings']['fs'] * pq.kHz
        elif params['recordings']['fs'] is None:
            params['recordings']['fs'] = 1. / temp_info['params']['dt']
            fs = params['recordings']['fs'] * pq.kHz
        else:
            fs = params['recordings']['fs'] * pq.kHz

        if 'noise_mode' not in rec_params.keys():
            params['recordings']['noise_mode'] = 'uncorrelated'
        noise_mode = params['recordings']['noise_mode']

        if 'noise_color' not in rec_params.keys():
            params['recordings']['noise_color'] = False
        noise_color = params['recordings']['noise_color']

        if 'sync_rate' not in rec_params.keys():
            params['recordings']['sync_rate'] = 0
        sync_rate = params['recordings']['sync_rate']

        if 'sync_jitt' not in rec_params.keys():
            params['recordings']['sync_jitt'] = 1
        sync_jitt = params['recordings']['sync_jitt'] * pq.ms

        if noise_mode == 'distance-correlated':
            if 'half_distance' not in rec_params.keys():
                params['recordings']['half_dist'] = 30
                half_dist = 30

        if noise_color is True:
            if 'color_peak' not in rec_params.keys():
                params['recordings']['color_peak'] = 500
            color_peak = params['recordings']['color_peak']
            if 'color_q' not in rec_params.keys():
                params['recordings']['color_q'] = 1
            color_q = params['recordings']['color_q']
            if 'random_noise_floor' not in rec_params.keys():
                params['recordings']['random_noise_floor'] = 0.5
            random_noise_floor = params['recordings']['random_noise_floor']

        if 'noise_level' not in rec_params.keys():
            params['recordings']['noise_level'] = 10
        noise_level = params['recordings']['noise_level']
        if self.verbose:
            print('Noise Level ', noise_level)

        if 'filter' not in rec_params.keys():
            params['recordings']['filter'] = True
        filter = params['recordings']['filter']

        if 'filter_cutoff' not in rec_params.keys():
            params['recordings']['filter_cutoff'] = [300., 6000.]
        cutoff = params['recordings']['filter_cutoff'] * pq.Hz

        if 'filter_order' not in rec_params.keys():
            params['recordings']['filter_order'] = 3
        order = params['recordings']['filter_order']

        if 'modulation' not in rec_params.keys():
            params['recordings']['modulation'] = 'electrode'
        modulation = params['recordings']['modulation']

        if 'shape_mod' not in rec_params.keys():
            params['recordings']['shape_mod'] = False
        shape_mod = params['recordings']['shape_mod']

        if shape_mod:
            if 'bursting_fc' not in rec_params.keys():
                params['recordings']['bursting_fc'] = [500., 12000.]
            bursting_fc = params['recordings']['bursting_fc']
            if self.verbose:
                print('Bursting with modulation frequencies: ', bursting_fc, ' Hz')
        else:
            bursting_fc = None

        if 'isi' in modulation:
            if 'exp_decay' not in rec_params.keys():
                params['recordings']['exp_decay'] = 0.2
            exp_decay = params['recordings']['exp_decay']
            if 'n_burst_spikes' not in rec_params.keys():
                params['recordings']['n_burst_spikes'] = 10
            n_burst_spikes = params['recordings']['n_burst_spikes']
            if 'max_burst_duration' not in rec_params.keys():
                params['recordings']['max_burst_duration'] = 100
            max_burst_duration = 100 * pq.ms

        if 'chunk_noise_duration' not in rec_params.keys():
            params['recordings']['chunk_noise_duration'] = 0
        chunk_noise_duration = params['recordings']['chunk_noise_duration'] * pq.s

        if 'chunk_filter_duration' not in rec_params.keys():
            params['recordings']['chunk_filter_duration'] = 0
        chunk_filter_duration = params['recordings']['chunk_filter_duration'] * pq.s

        if 'mrand' not in rec_params.keys():
            params['recordings']['mrand'] = 1
        mrand = params['recordings']['mrand']

        if 'sdrand' not in rec_params.keys():
            params['recordings']['sdrand'] = 0.05
        sdrand = params['recordings']['sdrand']

        if 'overlap' not in rec_params.keys():
            params['recordings']['overlap'] = False
        overlap = params['recordings']['overlap']

        if 'extract_waveforms' not in rec_params.keys():
            params['recordings']['extract_waveforms'] = False
        extract_waveforms = params['recordings']['extract_waveforms']

        if 'seed' not in rec_params.keys():
            params['recordings']['seed'] = np.random.randint(1, 1000)
        elif params['recordings']['seed'] is None:
            params['recordings']['seed'] = np.random.randint(1, 1000)
        noise_seed = params['recordings']['seed']

        if 'drifting' not in rec_params.keys():
            params['recordings']['drifting'] = False
        drifting = params['recordings']['drifting']

        if drifting:
            assert temp_info['params']['drifting']
            preferred_dir = np.array(rec_params['preferred_dir'])
            preferred_dir = preferred_dir / np.linalg.norm(preferred_dir)
            angle_tol = rec_params['angle_tol']
            drift_velocity = rec_params['drift_velocity']
            t_start_drift = rec_params['t_start_drift'] * pq.s
            if rec_params['n_drifting'] is None:
                n_drifting = n_neurons
            else:
                n_drifting = rec_params['n_drifting']
        else:
            # if drifting templates, but not recordings, consider initial template
            if temp_info['params']['drifting']:
                eaps = eaps[:, 0]
            preferred_dir = None
            angle_tol = None
            drift_velocity = None
            t_start_drift = None
            n_drifting = None

        if 'xlim' not in temp_params.keys():
            params['templates']['xlim'] = None
        x_lim = params['templates']['xlim']

        if 'ylim' not in temp_params.keys():
            params['templates']['ylim'] = None
        y_lim = params['templates']['ylim']

        if 'zlim' not in temp_params.keys():
            params['templates']['zlim'] = None
        z_lim = params['templates']['zlim']

        if 'min_amp' not in temp_params.keys():
            params['templates']['min_amp'] = 50
        min_amp = params['templates']['min_amp']

        if 'max_amp' not in temp_params.keys():
            params['templates']['max_amp'] = np.inf
        max_amp = params['templates']['max_amp']

        if 'min_dist' not in temp_params.keys():
            params['templates']['min_dist'] = 25
        min_dist = params['templates']['min_dist']

        if 'overlap_threshold' not in temp_params.keys():
            params['templates']['overlap_threshold'] = 0.8
        overlap_threshold = params['templates']['overlap_threshold']

        if 'pad_len' not in temp_params.keys():
            params['templates']['pad_len'] = [3., 3.]
        pad_len = params['templates']['pad_len']

        if 'n_jitters' not in temp_params.keys():
            params['templates']['n_jitters'] = 10
        n_jitters = params['templates']['n_jitters']

        if 'upsample' not in temp_params.keys():
            params['templates']['upsample'] = 8
        upsample = params['templates']['upsample']

        if 'seed' not in temp_params.keys():
            params['templates']['seed'] = np.random.randint(1, 1000)
        elif params['templates']['seed'] is None:
            params['templates']['seed'] = np.random.randint(1, 1000)
        temp_seed = params['templates']['seed']

        if celltype_params is not None:
            if 'excitatory' in celltype_params.keys() and 'inhibitory' in celltype_params.keys():
                exc_categories = celltype_params['excitatory']
                inh_categories = celltype_params['inhibitory']
                bin_cat = get_binary_cat(celltypes, exc_categories, inh_categories)
            else:
                bin_cat = np.array(['U'] * len(celltypes))
        else:
            bin_cat = np.array(['U'] * len(celltypes))

        # load MEA info
        electrode_name = temp_info['electrodes']['electrode_name']
        elinfo = mu.return_mea_info(electrode_name)
        offset = temp_info['params']['offset']
        elinfo['offset'] = offset
        mea = mu.return_mea(info=elinfo)
        mea_pos = mea.positions
        n_elec = mea_pos.shape[0]
        n_samples = int(duration.rescale('s').magnitude * fs.rescale('Hz').magnitude)

        params['recordings'].update({'duration': float(duration.magnitude),
                                     'fs': float(fs.rescale('Hz').magnitude),
                                     'n_neurons': n_neurons})
        params['electrodes'] = temp_info['electrodes']

        spike_duration = np.sum(temp_info['params']['cut_out']) * pq.ms
        spike_fs = 1. / temp_info['params']['dt'] * pq.kHz

        if self.verbose:
            print('Selecting cells')

        recordings = np.zeros((n_elec, n_samples))
        timestamps = np.arange(recordings.shape[1]) / fs

        if not only_noise:
            if 'type' in spiketrains[0].annotations.keys():
                n_exc = [st.annotations['type'] for st in spiketrains].count('E')
                n_inh = n_neurons - n_exc
            if self.verbose:
                print('Templates selection seed: ', temp_seed)
            np.random.seed(temp_seed)

            if drifting:
                drift_directions = np.array([(p[-1] - p[0]) / np.linalg.norm(p[-1] - p[0]) for p in locs])
                drift_velocity_ums = drift_velocity / 60.
                velocity_vector = drift_velocity_ums * preferred_dir
                if self.verbose:
                    print('Drift velocity vector: ', velocity_vector)
                n_elec = eaps.shape[2]
            else:
                drift_directions = None
                preferred_dir = None
                velocity_vector = None
                n_elec = eaps.shape[1]

            idxs_cells, selected_cat = select_templates(locs, eaps, bin_cat, n_exc, n_inh, x_lim=x_lim, y_lim=y_lim,
                                                        z_lim=z_lim, min_amp=min_amp, max_amp=max_amp,
                                                        min_dist=min_dist, drifting=drifting,
                                                        drift_dir=drift_directions,
                                                        preferred_dir=preferred_dir, angle_tol=angle_tol,
                                                        verbose=self.verbose)

            idxs_cells = np.array(idxs_cells)[np.argsort(selected_cat)]
            template_celltypes = celltypes[idxs_cells]
            template_locs = locs[idxs_cells]
            templates_bin = bin_cat[idxs_cells]
            templates = eaps[idxs_cells]

            # peak images
            voltage_peaks = []
            for tem in templates:
                dt = 1. / fs.magnitude
                if not drifting:
                    feat = get_templates_features(tem, ['na'], dt=dt)
                else:
                    feat = get_templates_features(tem[0], ['na'], dt=dt)
                voltage_peaks.append(-np.squeeze(feat['na']))
            voltage_peaks = np.array(voltage_peaks)

            # resample spikes
            up = fs
            down = spike_fs
            sampling_ratio = float(up / down)
            pad_samples = [int((pp * fs).magnitude) for pp in pad_len]
            n_resample = int((fs * spike_duration).magnitude)
            if not drifting:
                if templates.shape[2] != n_resample:
                    templates_pol = np.zeros((templates.shape[0], templates.shape[1], n_resample))
                    if self.verbose:
                        print('Resampling spikes')
                    for t, tem in enumerate(templates):
                        tem_pad = np.pad(tem, [(0, 0), pad_samples], 'edge')
                        tem_poly = ss.resample_poly(tem_pad, up, down, axis=1)
                        templates_pol[t, :] = tem_poly[:,
                                              int(sampling_ratio * pad_samples[0]):int(sampling_ratio * pad_samples[0])
                                                                                   + n_resample]
                else:
                    templates_pol = templates
            else:
                if templates.shape[3] != n_resample:
                    templates_pol = np.zeros((templates.shape[0], templates.shape[1], templates.shape[2], n_resample))
                    if self.verbose:
                        print('Resampling spikes')
                    for t, tem in enumerate(templates):
                        for ts, tem_single in enumerate(tem):
                            tem_pad = np.pad(tem_single, [(0, 0), pad_samples], 'edge')
                            tem_poly = ss.resample_poly(tem_pad, up, down, axis=1)
                            templates_pol[t, ts, :] = tem_poly[:,int(sampling_ratio *
                                                                     pad_samples[0]):int(sampling_ratio *
                                                                                         pad_samples[0]) + n_resample]
                else:
                    templates_pol = templates

            templates_pad = []
            if self.verbose:
                print('Padding template edges')
            for t, tem in enumerate(templates_pol):
                if not drifting:
                    tem = cubic_padding(tem, pad_len, fs)
                    templates_pad.append(tem)
                else:
                    if self.verbose:
                        print('Padding edges: neuron ', t + 1, ' of ', len(templates_pol))
                    templates_pad_p = []
                    for tem_p in tem:
                        tem_p = cubic_padding(tem_p, pad_len, fs)
                        templates_pad_p.append(tem_p)
                    templates_pad.append(templates_pad_p)
            templates_pad = np.array(templates_pad)

            if self.verbose:
                print('Creating time jittering')
            jitter = 1. / fs
            templates_jitter = []
            if not drifting:
                for temp in templates_pad:
                    temp_up = ss.resample_poly(temp, upsample, 1, axis=1)
                    nsamples_up = temp_up.shape[1]
                    temp_jitt = []
                    for n in np.arange(n_jitters):
                        # align waveform
                        shift = int((jitter * (np.random.random() - 0.5) * upsample * fs).magnitude)
                        if shift > 0:
                            t_jitt = np.pad(temp_up, [(0, 0), (np.abs(shift), 0)], 'constant')[:, :nsamples_up]
                        elif shift < 0:
                            t_jitt = np.pad(temp_up, [(0, 0), (0, np.abs(shift))], 'constant')[:, -nsamples_up:]
                        else:
                            t_jitt = temp_up
                        temp_down = ss.decimate(t_jitt, upsample, axis=1)
                        temp_jitt.append(temp_down)
                    templates_jitter.append(temp_jitt)
            else:
                for t, temp in enumerate(templates_pad):
                    if self.verbose:
                        print('Jittering: neuron ', t + 1, ' of ', len(templates_pol))
                    templates_jitter_p = []
                    for tem_p in temp:
                        temp_up = ss.resample_poly(tem_p, upsample, 1, axis=1)
                        nsamples_up = temp_up.shape[1]
                        temp_jitt = []
                        for n in np.arange(n_jitters):
                            # align waveform
                            shift = int((jitter * np.random.randn() * upsample * fs).magnitude)
                            if shift > 0:
                                t_jitt = np.pad(temp_up, [(0, 0), (np.abs(shift), 0)], 'constant')[:, :nsamples_up]
                            elif shift < 0:
                                t_jitt = np.pad(temp_up, [(0, 0), (0, np.abs(shift))], 'constant')[:, -nsamples_up:]
                            else:
                                t_jitt = temp_up
                            temp_down = ss.decimate(t_jitt, upsample, axis=1)
                            temp_jitt.append(temp_down)

                        templates_jitter_p.append(temp_jitt)
                    templates_jitter.append(templates_jitter_p)
            templates = np.array(templates_jitter)

            cut_outs_samples = np.array(cut_outs * fs.rescale('kHz').magnitude, dtype=int) + pad_samples

            del templates_pol, templates_pad, templates_jitter
            if drifting:
                del templates_jitter_p, templates_pad_p

            overlapping_computed = False
            if sync_rate != 0:
                if self.verbose:
                    print('Adding synchrony on overlapping spikes')
                overlapping = find_overlapping_templates(templates, thresh=overlap_threshold)
                overlapping_computed = True
                if self.verbose:
                    print('Overlapping templates: ', overlapping)
                for over in overlapping:
                    if self.verbose:
                        print('Overlapping pair: ', over)
                    spgen.add_synchrony(over, rate=sync_rate, verbose=self.verbose, time_jitt=sync_jitt)
                    # annotate new firing rates
                    fr1 = len(spgen.all_spiketrains[over[0]].times) / spgen.all_spiketrains[over[0]].t_stop
                    fr2 = len(spgen.all_spiketrains[over[1]].times) / spgen.all_spiketrains[over[1]].t_stop
                    spgen.all_spiketrains[over[0]].annotate(fr=fr1)
                    spgen.all_spiketrains[over[1]].annotate(fr=fr2)
            else:
                overlapping = []
            self.overlapping = overlapping

            # find SNR and annotate
            if self.verbose:
                print('Computing spike train SNR')
            for t_i, temp in enumerate(templates):
                min_peak = np.min(temp)
                snr = np.abs(min_peak / float(noise_level))
                spiketrains[t_i].annotate(snr=snr)

            if self.verbose:
                print('Adding spiketrain annotations')
            for i, st in enumerate(spiketrains):
                st.annotate(bintype=templates_bin[i], mtype=template_celltypes[i], soma_position=template_locs[i])
            if overlap:
                if not overlapping_computed:
                    if self.verbose:
                        print('Finding overlapping spikes')
                    overlapping = find_overlapping_templates(templates, thresh=overlap_threshold)
                annotate_overlapping_spikes(spiketrains, overlapping_pairs=overlapping, verbose=True)

            amp_mod = []
            cons_spikes = []

            if modulation == 'template':
                if self.verbose:
                    print('Template modulation')
                for st in spiketrains:
                    amp, cons = compute_modulation(st, mrand=mrand, sdrand=sdrand,
                                                   n_spikes=0)
                    amp_mod.append(amp)
                    cons_spikes.append(cons)
            elif modulation == 'electrode':
                if self.verbose:
                    print('Electrode modulaton')
                for st in spiketrains:
                    amp, cons = compute_modulation(st, n_el=n_elec, mrand=mrand, sdrand=sdrand,
                                                   n_spikes=0)
                    amp_mod.append(amp)
                    cons_spikes.append(cons)
            elif modulation == 'template-isi':
                if self.verbose:
                    print('Template-ISI modulation')
                for st in spiketrains:
                    amp, cons = compute_modulation(st, sdrand=sdrand,
                                                   n_spikes=n_burst_spikes, exp=exp_decay,
                                                   max_burst_duration=max_burst_duration)
                    amp_mod.append(amp)
                    cons_spikes.append(cons)
            elif modulation == 'electrode-isi':
                if self.verbose:
                    print('Electrode-ISI modulation')
                for st in spiketrains:
                    amp, cons = compute_modulation(st, n_el=n_elec, mrand=mrand, sdrand=sdrand,
                                                   n_spikes=n_burst_spikes, exp=exp_decay,
                                                   max_burst_duration=max_burst_duration)
                    amp_mod.append(amp)
                    cons_spikes.append(cons)

            spike_matrix = resample_spiketrains(spiketrains, fs=fs)
            n_samples = spike_matrix.shape[1]

            final_loc = []
            final_idxs = []

            # # modulated convolution
            spike_traces = []
            if drifting:
                drifting_units = np.random.permutation(n_neurons)[:n_drifting]
            else:
                drifting_units = []

            for st, spike_bin in enumerate(spike_matrix):
                if modulation == 'none':
                    # reset random seed to keep sampling of jitter spike same
                    seed = np.random.randint(10000)
                    np.random.seed(seed)

                    if drifting and st in drifting_units:
                        rec, fin_pos, final_idx, mix = convolve_drifting_templates_spiketrains(st, spike_bin,
                                                                                               templates[st],
                                                                                               cut_out=
                                                                                               cut_outs_samples,
                                                                                               fs=fs,
                                                                                               loc=
                                                                                               template_locs[st],
                                                                                               v_drift=
                                                                                               velocity_vector,
                                                                                               t_start_drift=
                                                                                               t_start_drift,
                                                                                               verbose=self.verbose)
                        recordings += rec
                        final_loc.append(fin_pos)
                        final_idxs.append(final_idx)
                        np.random.seed(seed)
                        spike_traces.append(convolve_single_template(st, spike_bin,
                                                                     templates[st, 0, :,
                                                                     np.argmax(voltage_peaks[st])],
                                                                     cut_out=cut_outs_samples))
                        spiketrains[st].annotate(drifting=True)
                        spiketrains[st].annotate(initial_soma_position=template_locs[st, 0])
                        spiketrains[st].annotate(final_soma_position=fin_pos)
                        spiketrains[st].annotate(final_idx=final_idx)
                    else:
                        if drifting:
                            template = templates[st, 0]
                        else:
                            template = templates[st]
                        recordings += convolve_templates_spiketrains(st, spike_bin, template,
                                                                     cut_out=cut_outs_samples, verbose=self.verbose)
                        np.random.seed(seed)
                        spike_traces.append(convolve_single_template(st, spike_bin,
                                                                     template[:, np.argmax(voltage_peaks[st])],
                                                                     cut_out=cut_outs_samples))
                        if drifting:
                            final_loc.append(template_locs[st, 0])
                            final_idxs.append(0)
                elif 'electrode' in modulation:
                    seed = np.random.randint(10000)
                    np.random.seed(seed)

                    if drifting and st in drifting_units:
                        rec, fin_pos, final_idx, mix = convolve_drifting_templates_spiketrains(st, spike_bin,
                                                                                               templates[st],
                                                                                               cut_out=
                                                                                               cut_outs_samples,
                                                                                               modulation=True,
                                                                                               mod_array=
                                                                                               amp_mod[st],
                                                                                               fs=fs,
                                                                                               loc=
                                                                                               template_locs[st],
                                                                                               v_drift=
                                                                                               velocity_vector,
                                                                                               t_start_drift=
                                                                                               t_start_drift,
                                                                                               bursting=shape_mod,
                                                                                               fc=bursting_fc,
                                                                                               verbose=self.verbose)
                        recordings += rec
                        final_loc.append(fin_pos)
                        final_idxs.append(final_idx)
                        np.random.seed(seed)
                        spike_traces.append(convolve_single_template(st, spike_bin,
                                                                     templates[st, 0, :,
                                                                     np.argmax(voltage_peaks[st])],
                                                                     cut_out=cut_outs_samples,
                                                                     modulation=True,
                                                                     mod_array=amp_mod[st][:,
                                                                               np.argmax(voltage_peaks[st])]))
                        spiketrains[st].annotate(drifting=True)
                        spiketrains[st].annotate(initial_soma_position=template_locs[st, 0])
                        spiketrains[st].annotate(final_soma_position=fin_pos)
                        spiketrains[st].annotate(final_idx=final_idx)
                    else:
                        if drifting:
                            template = templates[st, 0]
                        else:
                            template = templates[st]
                        recordings += convolve_templates_spiketrains(st, spike_bin, template,
                                                                     cut_out=cut_outs_samples,
                                                                     modulation=True,
                                                                     mod_array=amp_mod[st],
                                                                     bursting=shape_mod,
                                                                     fc=bursting_fc,
                                                                     fs=fs, verbose=self.verbose)
                        np.random.seed(seed)
                        spike_traces.append(convolve_single_template(st, spike_bin,
                                                                     template[:, np.argmax(voltage_peaks[st])],
                                                                     cut_out=cut_outs_samples,
                                                                     modulation=True,
                                                                     mod_array=amp_mod[st][:,
                                                                               np.argmax(voltage_peaks[st])],
                                                                     bursting=shape_mod,
                                                                     fc=bursting_fc,
                                                                     fs=fs))
                        if drifting:
                            final_loc.append(template_locs[st, 0])
                            final_idxs.append(0)
                elif 'template' in modulation:
                    seed = np.random.randint(10000)
                    np.random.seed(seed)
                    if drifting and st in drifting_units:
                        rec, fin_pos, final_idx, mix = convolve_drifting_templates_spiketrains(st, spike_bin,
                                                                                               templates[st],
                                                                                               cut_out=
                                                                                               cut_outs_samples,
                                                                                               modulation=True,
                                                                                               mod_array=
                                                                                               amp_mod[st],
                                                                                               fs=fs,
                                                                                               loc=
                                                                                               template_locs[st],
                                                                                               v_drift=
                                                                                               velocity_vector,
                                                                                               t_start_drift=
                                                                                               t_start_drift,
                                                                                               bursting=shape_mod,
                                                                                               fc=bursting_fc,
                                                                                               verbose=self.verbose)
                        recordings += rec
                        final_idxs.append(final_idx)
                        final_loc.append(fin_pos)
                        np.random.seed(seed)
                        spike_traces.append(convolve_single_template(st, spike_bin,
                                                                     templates[st, 0, :,
                                                                     np.argmax(voltage_peaks[st])],
                                                                     cut_out=cut_outs_samples,
                                                                     modulation=True,
                                                                     mod_array=amp_mod[st]))
                        spiketrains[st].annotate(drifting=True)
                        spiketrains[st].annotate(initial_soma_position=template_locs[st, 0])
                        spiketrains[st].annotate(final_soma_position=fin_pos)
                        spiketrains[st].annotate(final_idx=final_idx)
                    else:
                        if drifting:
                            template = templates[st, 0]
                        else:
                            template = templates[st]
                        recordings += convolve_templates_spiketrains(st, spike_bin, template,
                                                                     cut_out=cut_outs_samples,
                                                                     modulation=True,
                                                                     mod_array=amp_mod[st],
                                                                     bursting=shape_mod,
                                                                     fc=bursting_fc,
                                                                     fs=fs, verbose=self.verbose)
                        np.random.seed(seed)
                        spike_traces.append(convolve_single_template(st, spike_bin,
                                                                     template[:, np.argmax(voltage_peaks[st])],
                                                                     cut_out=cut_outs_samples,
                                                                     modulation=True,
                                                                     mod_array=amp_mod[st],
                                                                     bursting=shape_mod,
                                                                     fc=bursting_fc,
                                                                     fs=fs))
                        if drifting:
                            final_loc.append(template_locs[st, 0])
                            final_idxs.append(0)
            spike_traces = np.array(spike_traces)

            if drifting:
                templates_drift = np.zeros((templates.shape[0], final_idx + 1,
                                            templates.shape[2], templates.shape[3],
                                            templates.shape[4]))
                for i, st in enumerate(spiketrains):
                    if i in drifting_units:
                        templates_drift[i] = templates[i, :(final_idx + 1)]
                    else:
                        templates_drift[i] = np.array([templates[i, 0]] * (final_idx + 1))
                templates = templates_drift
        else:
            spiketrains = np.array([])
            voltage_peaks = np.array([])
            spike_traces = np.array([])
            templates = np.array([])
            overlapping = np.array([])

        if self.verbose:
            print('Adding noise')
        # divide in chunks
        chunks_noise = []
        if duration > chunk_noise_duration and chunk_noise_duration != 0:
            start = 0 * pq.s
            finished = False
            while not finished:
                chunks_noise.append([start, start + chunk_noise_duration])
                start = start + chunk_noise_duration
                if start >= duration:
                    finished = True

        if self.verbose:
            print('Noise seed: ', noise_seed)
        if noise_color:
            if self.verbose:
                print('Coloring noise with peak: ', color_peak, ' quality factor: ', color_q,
                      ' and random noise level: ', random_noise_floor)
        np.random.seed(noise_seed)
        if noise_level > 0:
            if noise_mode == 'uncorrelated':
                if len(chunks_noise) > 0:
                    for ch, chunk in enumerate(chunks_noise):
                        if self.verbose:
                            print('Generating noise in: ', chunk[0], chunk[1], ' chunk')
                        idxs = np.where((timestamps >= chunk[0]) & (timestamps < chunk[1]))[0]
                        additive_noise = noise_level * np.random.randn(recordings.shape[0],
                                                                       len(idxs))
                        if noise_color:
                            # iir peak filter
                            b_iir, a_iir = ss.iirpeak(color_peak, Q=color_q, fs=fs.rescale('Hz').magnitude)
                            additive_noise = ss.filtfilt(b_iir, a_iir, additive_noise, axis=1, padlen=1000)
                            additive_noise = additive_noise + random_noise_floor * np.std(additive_noise) * \
                                             np.random.randn(additive_noise.shape[0], additive_noise.shape[1])
                            additive_noise = additive_noise * (noise_level / np.std(additive_noise))
                        recordings[:, idxs] += additive_noise
                else:
                    additive_noise = noise_level * np.random.randn(recordings.shape[0],
                                                                   recordings.shape[1])
                    if noise_color:
                        # iir peak filter
                        b_iir, a_iir = ss.iirpeak(color_peak, Q=color_q, fs=fs.rescale('Hz').magnitude)
                        additive_noise = ss.filtfilt(b_iir, a_iir, additive_noise, axis=1)
                        additive_noise = additive_noise + random_noise_floor * np.std(additive_noise) \
                                         * np.random.randn(additive_noise.shape[0], additive_noise.shape[1])
                        additive_noise = additive_noise * (noise_level / np.std(additive_noise))

                    recordings += additive_noise
            elif noise_mode == 'distance-correlated':
                cov_dist = np.zeros((n_elec, n_elec))
                for i, el in enumerate(mea.positions):
                    for j, p in enumerate(mea.positions):
                        if i != j:
                            cov_dist[i, j] = (0.5 * half_dist) / np.linalg.norm(el - p)
                        else:
                            cov_dist[i, j] = 1

                if len(chunks_noise) > 0:
                    for ch, chunk in enumerate(chunks_noise):
                        if self.verbose:
                            print('Generating noise in: ', chunk[0], chunk[1], ' chunk')
                        idxs = np.where((timestamps >= chunk[0]) & (timestamps < chunk[1]))[0]
                        additive_noise = noise_level * np.random.multivariate_normal(np.zeros(n_elec), cov_dist,
                                                                                     size=(len(idxs))).T
                        if noise_color:
                            # iir peak filter
                            b_iir, a_iir = ss.iirpeak(color_peak, Q=color_q, fs=fs.rescale('Hz').magnitude)
                            additive_noise = ss.filtfilt(b_iir, a_iir, additive_noise, axis=1)
                            additive_noise = additive_noise + random_noise_floor * np.std(additive_noise) \
                                             * np.random.randn(additive_noise.shape[0], additive_noise.shape[1])
                            additive_noise = additive_noise * (noise_level / np.std(additive_noise))
                        recordings[:, idxs] += additive_noise
                else:
                    additive_noise = noise_level * np.random.multivariate_normal(np.zeros(n_elec), cov_dist,
                                                                                 size=recordings.shape[1]).T
                    if noise_color:
                        # iir peak filter
                        b_iir, a_iir = ss.iirpeak(color_peak, Q=color_q, fs=fs.rescale('Hz').magnitude)
                        additive_noise = ss.filtfilt(b_iir, a_iir, additive_noise, axis=1)
                        additive_noise = additive_noise + random_noise_floor * np.std(additive_noise) \
                                         * np.random.randn(additive_noise.shape[0], additive_noise.shape[1])
                        additive_noise = additive_noise * (noise_level / np.std(additive_noise))

                    recordings += additive_noise
        else:
            if self.verbose:
                print('Noise level is set to 0')

        if filter:
            if self.verbose:
                print('Filtering')
                if cutoff.size == 1:
                    print('High-pass cutoff', cutoff)
                elif cutoff.size == 2:
                    print('Band-pass cutoff', cutoff)
            chunks_filter = []
            if duration > chunk_filter_duration and chunk_filter_duration != 0:
                start = 0 * pq.s
                finished = False
                while not finished:
                    chunks_filter.append([start, start + chunk_filter_duration])
                    start = start + chunk_filter_duration
                    if start >= duration:
                        finished = True
            if len(chunks_filter) > 0:
                for ch, chunk in enumerate(chunks_filter):
                    if self.verbose:
                        print('Filtering in: ', chunk[0], chunk[1], ' chunk')
                    idxs = np.where((timestamps >= chunk[0]) & (timestamps < chunk[1]))[0]
                    if cutoff.size == 1:
                        recordings[:, idxs] = filter_analog_signals(recordings[:, idxs], freq=cutoff, fs=fs,
                                                                    filter_type='highpass', order=order)
                    elif cutoff.size == 2:
                        if fs / 2. < cutoff[1]:
                            recordings[:, idxs] = filter_analog_signals(recordings[:, idxs], freq=cutoff[0], fs=fs,
                                                                        filter_type='highpass', order=order)
                        else:
                            recordings[:, idxs] = filter_analog_signals(recordings[:, idxs], freq=cutoff, fs=fs)
            else:
                if cutoff.size == 1:
                    recordings = filter_analog_signals(recordings, freq=cutoff, fs=fs, filter_type='highpass',
                                                       order=order)
                elif cutoff.size == 2:
                    if fs / 2. < cutoff[1]:
                        recordings = filter_analog_signals(recordings, freq=cutoff[0], fs=fs,
                                                           filter_type='highpass', order=order)
                    else:
                        recordings = filter_analog_signals(recordings, freq=cutoff, fs=fs, order=order)

        if not only_noise:
            if extract_waveforms:
                if self.verbose:
                    print('Extracting spike waveforms')
                extract_wf(spiketrains, recordings, fs=fs, timestamps=timestamps)

        params['templates']['overlapping'] = np.array(overlapping)
        self.recordings = recordings
        self.timestamps = timestamps
        self.channel_positions = mea_pos
        self.templates = np.squeeze(templates)
        self.spiketrains = spiketrains
        self.voltage_peaks = voltage_peaks
        self.spike_traces = spike_traces
        self.info = params

    def annotate_overlapping_spike(self, parallel=True):
        '''
        Annnotate spike trains with overlapping information.
        '''
        annotate_overlapping_spikes(self.spiketrains, overlapping_pairs=self.overlapping, parallel=parallel)


def gen_recordings(params=None, templates=None, tempgen=None, verbose=True):
    '''
    Generates recordings.

    Parameters
    ----------
    templates : str
        Path to generated templates
    params : dict OR str
        Dictionary containing recording parameters OR path to yaml file containing parameters
    tempgen : TemplateGenerator
        Template generator object
    verbose : bool
        If True output is verbose

    Returns
    -------
    RecordingGenerator
        Generated recording generator object

    '''
    t_start = time.time()
    if isinstance(params, str):
        if os.path.isfile(params) and (params.endswith('yaml') or params.endswith('yml')):
            with open(params, 'r') as pf:
                if use_loader:
                    params_dict = yaml.load(pf, Loader=yaml.FullLoader)
                else:
                    params_dict = yaml.load(pf)
    elif isinstance(params, dict):
        params_dict = params
    else:
        params_dict = {}

    if 'spiketrains' not in params_dict:
        params_dict['spiketrains'] = {}
    if 'templates' not in params_dict:
        params_dict['templates'] = {}
    if 'recordings' not in params_dict:
        params_dict['recordings'] = {}
    if 'cell_types' not in params_dict:
        params_dict['cell_types'] = {}

    if tempgen is None and templates is None:
        raise AttributeError("Provide either 'templates' or 'tempgen' TemplateGenerator object")

    if tempgen is None:
        if os.path.isdir(templates):
            tempgen = load_templates(templates, verbose=False)
        elif templates.endswith('h5') or templates.endswith('hdf5'):
            tempgen = load_templates(templates, verbose=False)
        else:
            raise AttributeError("'templates' is not a folder or an hdf5 file")

    if 'seed' in params_dict['spiketrains']:
        if params_dict['spiketrains']['seed'] is None:
            params_dict['spiketrains']['seed'] = np.random.randint(1, 10000)
    else:
        params_dict['spiketrains'].update({'seed': np.random.randint(1, 10000)})

    if 'seed' in params_dict['templates']:
        if params_dict['templates']['seed'] is None:
            params_dict['templates']['seed'] = np.random.randint(1, 10000)
    else:
        params_dict['templates'].update({'seed': np.random.randint(1, 10000)})

    if 'seed' in params_dict['recordings']:
        if params_dict['recordings']['seed'] is None:
            params_dict['recordings']['seed'] = np.random.randint(1, 10000)
    else:
        params_dict['recordings'].update({'recordings': np.random.randint(1, 10000)})

    # Generate spike trains
    spgen = SpikeTrainGenerator(params_dict['spiketrains'], verbose=verbose)
    spgen.generate_spikes()
    spiketrains = spgen.all_spiketrains

    params_dict['spiketrains'] = spgen.info
    # Generate recordings
    recgen = RecordingGenerator(spgen, tempgen, params_dict, verbose=verbose)
    recgen.generate_recordings()

    print('Elapsed time: ', time.time() - t_start)

    return recgen


def gen_templates(cell_models_folder, params=None, templates_folder=None,
                  intraonly=False, parallel=True, delete_tmp=True, verbose=True):
    '''

    Parameters
    ----------
    tmp_params_path : str
        path to tmp params.yaml (same as templates_params with extra 'simulate_script', 'templates_folder' fields)
    intraonly : bool
        if True only intracellular simulation is run
    parallel : bool
        if True multi-threading is used
    delete_tmp :
        if True the temporary files are deleted
    verbose : bool
        If True output is verbose

    Returns
    -------
    TemplateGenerator
        Generated template generator object

    '''
    if isinstance(params, str):
        if os.path.isfile(params) and (params.endswith('yaml') or params.endswith('yml')):
            with open(params, 'r') as pf:
                if use_loader:
                    params_dict = yaml.load(pf, Loader=yaml.FullLoader)
                else:
                    params_dict = yaml.load(pf)
    elif isinstance(params, dict):
        params_dict = params
    else:
        params_dict = None

    if templates_folder is not None:
        if not os.path.isdir(templates_folder):
            os.makedirs(templates_folder)

    tempgen = TemplateGenerator(cell_models_folder=cell_models_folder,
                                params=params_dict,
                                templates_folder=templates_folder,
                                intraonly=intraonly,
                                parallel=parallel,
                                delete_tmp=delete_tmp,
                                verbose=verbose)
    tempgen.generate_templates()

    return tempgen



from MEArec.tools import load_templates, load_recordings, save_recording_generator, save_template_generator, \
    get_default_config, plot_templates, plot_recordings, plot_rasters, plot_waveforms, get_templates_features
from MEArec.generators import gen_recordings, gen_templates, RecordingGenerator, TemplateGenerator, SpikeTrainGenerator


__version__ = '1.2.4'



#!/usr/bin/env python

"""Python script to run cell model"""


"""
/* Copyright (c) 2015 EPFL-BBP, All rights reserved.

THIS SOFTWARE IS PROVIDED BY THE BLUE BRAIN PROJECT ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE BLUE BRAIN PROJECT
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This work is licensed under a
Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
To view a copy of this license, visit
http://creativecommons.org/licenses/by-nc-sa/4.0/legalcode or send a letter to
Creative Commons, 171 Second Street, Suite 300,
San Francisco, California, 94105, USA.
"""

"""
 * @file run.py
 * @brief Run simulation using pyneuron
 * @author Werner Van Geit @ BBP
 * @date 2015
"""

# pylint: disable=C0325, W0212, F0401, W0612, F0401

import os
import neuron
import numpy
import sys


def create_cell(add_synapses=True):
    """Create the cell model"""
    # Load morphology
    neuron.h.load_file("morphology.hoc")
    # Load biophysics
    neuron.h.load_file("biophysics.hoc")
    # Load main cell template
    neuron.h.load_file("template.hoc")

    # Instantiate the cell from the template

    print("Loading cell bAC217_L5_DBC_4765d943f4")
    cell = neuron.h.bAC217_L5_DBC_4765d943f4(1 if add_synapses else 0)
    return cell


def create_stimuli(cell, step_number):
    """Create the stimuli"""

    print('Attaching stimulus electrodes')

    stimuli = []
    step_amp = [0] * 3

    with open('current_amps.dat', 'r') as current_amps_file:
        first_line = current_amps_file.read().split('\n')[0].strip()
        hyp_amp, step_amp[0], step_amp[1], step_amp[2] = first_line.split(' ')

    iclamp = neuron.h.IClamp(0.5, sec=cell.soma[0])
    iclamp.delay = 700
    iclamp.dur = 2000
    iclamp.amp = float(step_amp[step_number - 1])
    print('Setting up step current clamp: '
          'amp=%f nA, delay=%f ms, duration=%f ms' %
          (iclamp.amp, iclamp.delay, iclamp.dur))

    stimuli.append(iclamp)

    hyp_iclamp = neuron.h.IClamp(0.5, sec=cell.soma[0])
    hyp_iclamp.delay = 0
    hyp_iclamp.dur = 3000
    hyp_iclamp.amp = float(hyp_amp)
    print('Setting up hypamp current clamp: '
          'amp=%f nA, delay=%f ms, duration=%f ms' %
          (hyp_iclamp.amp, hyp_iclamp.delay, hyp_iclamp.dur))

    stimuli.append(hyp_iclamp)

    return stimuli


def create_recordings(cell):
    """Create the recordings"""
    print('Attaching recording electrodes')

    recordings = {}

    recordings['time'] = neuron.h.Vector()
    recordings['soma(0.5)'] = neuron.h.Vector()

    recordings['time'].record(neuron.h._ref_t, 0.1)
    recordings['soma(0.5)'].record(cell.soma[0](0.5)._ref_v, 0.1)

    return recordings


def run_step(step_number, plot_traces=None):
    """Run step current simulation with index step_number"""

    cell = create_cell(add_synapses=False)
    stimuli = create_stimuli(cell, step_number)
    recordings = create_recordings(cell)

    # Overriding default 30s simulation,
    print('Setting simulation time to 3s for the step currents')
    neuron.h.tstop = 3000

    print('Disabling variable timestep integration')
    neuron.h.cvode_active(0)

    print('Running for %f ms' % neuron.h.tstop)
    neuron.h.run()

    time = numpy.array(recordings['time'])
    soma_voltage = numpy.array(recordings['soma(0.5)'])

    recordings_dir = 'python_recordings'

    soma_voltage_filename = os.path.join(
        recordings_dir,
        'soma_voltage_step%d.dat' % step_number)
    numpy.savetxt(
            soma_voltage_filename,
            numpy.transpose(
               numpy.vstack((
                    time,
                    soma_voltage))))

    print('Soma voltage for step %d saved to: %s'
          % (step_number, soma_voltage_filename))

    if plot_traces:
        import pylab
        pylab.figure()
        pylab.plot(recordings['time'], recordings['soma(0.5)'])
        pylab.xlabel('time (ms)')
        pylab.ylabel('Vm (mV)')
        pylab.gcf().canvas.set_window_title('Step %d' % step_number)


def init_simulation():
    """Initialise simulation environment"""

    neuron.h.load_file("stdrun.hoc")
    neuron.h.load_file("import3d.hoc")

    print('Loading constants')
    neuron.h.load_file('constants.hoc')


def main(plot_traces=True):
    """Main"""

    # Import matplotlib to plot the traces
    if plot_traces:
        import matplotlib
        matplotlib.rcParams['path.simplify'] = False

    init_simulation()

    for step_number in range(1, 4):
        run_step(step_number, plot_traces=plot_traces)

    if plot_traces:
        import pylab
        pylab.show()

if __name__ == '__main__':
    if len(sys.argv) == 1:
        main(plot_traces=True)
    elif len(sys.argv) == 2 and sys.argv[1] == '--no-plots':
        main(plot_traces=False)
    else:
        raise Exception(
            "Script only accepts one argument: --no-plots, not %s" %
            str(sys.argv))



import os
from os.path import join
import time
import MEAutility as mu
import click
import numpy as np
import yaml
import shutil
import MEArec.generators as generators
from MEArec import save_template_generator, save_recording_generator, get_default_config
import pprint
import time
from distutils.version import StrictVersion

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False

@click.group()
def cli():
    """MEArec: Fast and customizable simulation of extracellular recordings on Multi-Electrode-Arrays """
    pass


@cli.command()
@click.option('--params', '-prm', default=None,
              help='path to default_params.yaml (otherwise default default_params are used and some of the parameters'
                   'can be overwritten with the following options)')
@click.option('--default', is_flag=True,
              help='shows default values for simulation')
@click.option('--fname', '-fn', default=None,
              help='template filename')
@click.option('--folder', '-fol', default=None,
              help='templates output base folder')
@click.option('--cellfolder', '-cf', default=None,
              help='folder containing bbp cell models')
@click.option('--rot', '-r', default=None,
              help='possible rotation arguments: norot-xrot-yrot-zrot-physrot-3drot (default=physrot)')
@click.option('--probe', '-prb', default=None,
              help='probe name from available electrodes (default=None)')
@click.option('--n', '-n', default=None, type=int,
              help='number of observations per cell type (default=1000)')
@click.option('--ncontacts', '-nc', default=None, type=int,
              help='number of contacts per electrode (default=1)')
@click.option('--overhang', '-ov', default=None, type=float,
              help='extension (um) beyond MEA boundaries for neuron locations (default=30.)')
@click.option('--offset', '-off', default=None, type=float,
              help='x_plane (um) coordinate for MEA (default=0)')
@click.option('--xlim', '-xl', default=None, nargs=2, type=float,
              help='limits ( low high ) for neuron locations in the x-axis (depth) (default=[10.,80.])')
@click.option('--ylim', '-yl', default=None, nargs=2, type=float,
              help='limits ( low high ) for neuron locations in the y-axis (default=None)')
@click.option('--zlim', '-zl', default=None, nargs=2, type=float,
              help='limits ( low high ) for neuron locations in the z-axis (default=None)')
@click.option('--det-thresh', '-dt', default=None, type=float,
              help='detection threshold for EAPs (default=30)')
@click.option('--seed', '-s', default=None, type=int,
              help='random seed for template generation (int)')
@click.option('--intraonly', '-io', is_flag=True,
              help='only run intracellular simulations')
@click.option('--no-parallel', '-nopar', is_flag=True,
              help='run without multiprocessing tool')
@click.option('--drifting', '-dr', is_flag=True,
              help='generate drifting templates')
@click.option('--min-drift', '-mind', type=float,
              help='miniumum drifting distance')
@click.option('--max-drift', '-maxd', type=float,
              help='maximum drifting distance')
@click.option('--drift-steps', '-drst', type=int,
              help='number of drift steps')
@click.option('--drift-xlim', '-dxl', default=None, nargs=2, type=float,
              help='limits ( low high ) for neuron drift locations in the x-axis (depth) (default=[10.,80.])')
@click.option('--drift-ylim', '-dyl', default=None, nargs=2, type=float,
              help='limits ( low high ) for neuron drift locations in the y-axis (default=None)')
@click.option('--drift-zlim', '-dzl', default=None, nargs=2, type=float,
              help='limits ( low high ) for neuron drift locations in the z-axis (default=None)')
@click.option('--verbose', '-v', is_flag=True,
              help='produce verbose output')
def gen_templates(params, **kwargs):
    """Generates EAP templates on multi-electrode arrays using biophyical NEURON simulations and LFPy"""
    info, config_folder = get_default_config()

    if params is None:
        with open(info['templates_params'], 'r') as pf:
            if use_loader:
                params_dict = yaml.load(pf, Loader=yaml.FullLoader)
            else:
                params_dict = yaml.load(pf)
    else:
        with open(params, 'r') as pf:
            if use_loader:
                params_dict = yaml.load(pf, Loader=yaml.FullLoader)
            else:
                params_dict = yaml.load(pf)

    if kwargs['default'] is True:
        pprint.pprint(params_dict)
        mu.return_mea()
        return

    if kwargs['cellfolder'] is not None:
        model_folder = kwargs['cellfolder']
    else:
        model_folder = info['cell_models_folder']
    params_dict['cell_models_folder'] = model_folder

    if kwargs['seed'] is not None:
        seed = kwargs['seed']
    else:
        seed = np.random.randint(1, 10000)
    params_dict['seed'] = seed

    if kwargs['folder'] is not None:
        templates_folder = kwargs['folder']
        params_dict['templates_folder'] = kwargs['folder']
    else:
        templates_folder = info['templates_folder']
    intraonly = kwargs['intraonly']

    if kwargs['rot'] is not None:
        params_dict['rot'] = kwargs['rot']
    if kwargs['n'] is not None:
        params_dict['n'] = kwargs['n']
    if kwargs['ncontacts'] is not None:
        params_dict['ncontacts'] = kwargs['ncontacts']
    if kwargs['overhang'] is not None:
        params_dict['overhang'] = kwargs['overhang']
    if kwargs['offset'] is not None:
        params_dict['offset'] = kwargs['offset']
    if kwargs['det_thresh'] is not None:
        params_dict['det_thresh'] = kwargs['det_thresh']
    if kwargs['xlim'] is not None and len(kwargs['xlim']) == 2:
        params_dict['xlim'] = kwargs['xlim']
    if kwargs['ylim'] is not None and len(kwargs['ylim']) == 2:
        params_dict['ylim'] = kwargs['ylim']
    if kwargs['zlim'] is not None and len(kwargs['zlim']) == 2:
        params_dict['zlim'] = kwargs['zlim']

    if kwargs['drifting']:
        params_dict['drifting'] = True
    else:
        params_dict['drifting'] = False
    if kwargs['min_drift'] is not None:
        params_dict['min_drift'] = kwargs['min_drift']
    if kwargs['max_drift'] is not None:
        params_dict['max_drift'] = kwargs['max_drift']
    if kwargs['drift_steps'] is not None:
        params_dict['drift_steps'] = kwargs['drift_steps']
    if kwargs['drift_xlim'] is not None and len(kwargs['drift_xlim']) == 2:
        params_dict['drift_xlim'] = kwargs['drift_xlim']
    if kwargs['drift_ylim'] is not None and len(kwargs['drift_ylim']) == 2:
        params_dict['drift_ylim'] = kwargs['drift_ylim']
    if kwargs['drift_zlim'] is not None and len(kwargs['drift_zlim']) == 2:
        params_dict['drift_zlim'] = kwargs['drift_zlim']

    if kwargs['probe'] is not None:
        params_dict['probe'] = kwargs['probe']
    if kwargs['no_parallel']:
        parallel = False
    else:
        parallel = True
    verbose = kwargs['verbose']

    params_dict['templates_folder'] = templates_folder

    tempgen = generators.gen_templates(cell_models_folder=model_folder,
                                       params=params_dict,
                                       templates_folder=templates_folder,
                                       intraonly=intraonly,
                                       parallel=parallel,
                                       verbose=verbose)

    # Merge simulated data and cleanup
    if not intraonly:
        rot = params_dict['rot']
        n = params_dict['n']
        probe = params_dict['probe']
        if kwargs['fname'] is None:
            if params_dict['drifting']:
                fname = 'templates_%d_%s_drift_%s.h5' % (n, probe, time.strftime("%d-%m-%Y"))
            else:
                fname = 'templates_%d_%s_%s.h5' % (n, probe, time.strftime("%d-%m-%Y"))
        else:
            fname = kwargs['fname']
        save_fname = join(templates_folder, rot, fname)
        save_template_generator(tempgen, save_fname)


@cli.command()
@click.option('--templates', '-t', default=None,
              help='eap templates path')
@click.option('--params', '-prm', default=None,
              help='path to default_params.yaml (otherwise default default_params are used and some of the parameters can be overwritten with the following options)')
@click.option('--default', is_flag=True,
              help='shows default values for simulation')
@click.option('--fname', '-fn', default=None,
              help='recording filename')
@click.option('--folder', '-fol', default=None,
              help='recording output base folder')
@click.option('--duration', '-d', default=None, type=float,
              help='duration in s (default=10)')
@click.option('--n-exc', '-ne', default=None, type=int,
              help='number of excitatory cells (default=15)')
@click.option('--n-inh', '-ni', default=None, type=int,
              help='number of inhibitory cells (default=5)')
@click.option('--f-exc', '-fe', default=None, type=float,
              help='average firing rate of excitatory cells in Hz (default=5)')
@click.option('--f-inh', '-fi', default=None, type=float,
              help='average firing rate of inhibitory cells in Hz (default=15)')
@click.option('--st-exc', '-se', default=None, type=float,
              help='firing rate standard deviation of excitatory cells in Hz (default=1)')
@click.option('--st-inh', '-si', default=None, type=float,
              help='firing rate standard deviation of inhibitory cells in Hz (default=3)')
@click.option('--min-rate', '-mr', default=None, type=float,
              help='minimum firing rate (default=0.5)')
@click.option('--ref-per', '-rp', default=None, type=float,
              help='refractory period in ms (default=2)')
@click.option('--process', '-p', default='poisson', type=click.Choice(['poisson', 'gamma']),
              help='process for generating spike trains (default=poisson)')
@click.option('--tstart', default=None, type=float,
              help='start time in s (default=0)')
@click.option('--min-dist', '-md', default=None, type=int,
              help='minumum distance between neuron in um (default=25)')
@click.option('--min-amp', '-mina', default=None, type=int,
              help='minumum eap amplitude in uV (default=50)')
@click.option('--max-amp', '-maxa', default=None, type=int,
              help='maximum eap amplitude in uV (default=inf)')
@click.option('--fs', default=None, type=float,
              help='sampling frequency in kHz (default from templates sampling frequency)')
@click.option('--sync-rate', '-sr', default=None, type=float,
              help='added synchrony rate on spatially overlapping spikes')
@click.option('--sync-jitt', '-sj', default=None, type=float,
              help='jitter in ms for added spikes')
@click.option('--noise-lev', '-nl', default=None, type=int,
              help='noise level in uV (default=10)')
@click.option('--modulation', '-m', default=None, type=click.Choice(['none', 'template', 'electrode']),
              help='modulation type')
@click.option('--chunk-noise', '-chn', default=None, type=float,
              help='chunk duration in s for chunk noise (default 0)')
@click.option('--chunk-filt', '-chf', default=None, type=float,
              help='chunk duration in s for chunk filter (default 0)')
@click.option('--noise-seed', '-nseed', default=None, type=int,
              help='random seed for noise')
@click.option('--half-dist', '-hd', default=None, type=float,
              help='when noise is distance-correlated the distance at which covariance is 0.5 (default=30)')
@click.option('--color-noise', '-cn', is_flag=True,
              help='if True noise is colored')
@click.option('--color-peak', '-cp', default=None, type=float,
              help='peak for noise in Hz (default 500 Hz)')
@click.option('--color-q', '-cq', default=None, type=float,
              help='quality factor for noise filter (default=1)')
@click.option('--random-noise-floor', '-rnf', default=None, type=float,
              help='noise floor in std of additive noide (default 1)')
@click.option('--st-seed', '-stseed', default=None, type=int,
              help='random seed for spike trains')
@click.option('--temp-seed', '-tseed', default=None, type=int,
              help='random seed for template selection')
@click.option('--no-filt', is_flag=True,
              help='if True no filter is applied')
@click.option('--filt-cutoff', '-fc', default=None, type=float, multiple=True,
              help='filter cutoff frequencies.'
                   'High-pass: -fc hp-cutoff. Band-pass: -fc hp-cutoff -fc lp-cutoff')
@click.option('--filt-order', '-fo', default=None, type=int,
              help='filter order (default 3)')
@click.option('--overlap', is_flag=True,
              help='if True it annotates overlapping spikes')
@click.option('--overlap-thresh', '-ot', type=float,
              help='overlap threshold for spatial overlap')
@click.option('--extract-wf', is_flag=True,
              help='if True it annotates overlapping spikes')
@click.option('--drifting', '-dr', is_flag=True,
              help='generate drifting recordings')
@click.option('--preferred-dir', '-prd', default=None, nargs=3, type=float,
              help='preferred drifting direction (0 0 1 is positive z-direction)')
@click.option('--angle-tol', '-angt', type=float,
              help='angle tollerance for preferred direction')
@click.option('--drift-velocity', '-drvel', type=float,
              help='drift velocity in um/min')
@click.option('--t-start-drift', '-tsd', type=float,
              help='drifting start time in s')
@click.option('--verbose', '-v', is_flag=True,
              help='produce verbose output')
def gen_recordings(params, **kwargs):
    """Generates recordings from TEMPLATES"""
    # Retrieve default_params file
    info, config_folder = get_default_config()

    if params is None:
        with open(info['recordings_params'], 'r') as pf:
            if use_loader:
                params_dict = yaml.load(pf, Loader=yaml.FullLoader)
            else:
                params_dict = yaml.load(pf)
    else:
        with open(params, 'r') as pf:
            if use_loader:
                params_dict = yaml.load(pf, Loader=yaml.FullLoader)
            else:
                params_dict = yaml.load(pf)

    if kwargs['default'] is True:
        pprint.pprint(params_dict)
        return

    if kwargs['folder'] is not None:
        params_dict['recordings'].update({'recordings_folder': kwargs['folder']})
    else:
        params_dict['recordings'].update({'recordings_folder': info['recordings_folder']})
    recordings_folder = params_dict['recordings']['recordings_folder']

    if kwargs['templates'] is None:
        print('Provide eap templates path')
        return
    else:
        if os.path.isdir(kwargs['templates']) or kwargs['templates'].endswith('h5') \
                or kwargs['templates'].endswith('hdf5'):
            params_dict['templates'].update({'templates': kwargs['templates']})
        else:
            print(kwargs['templates'])
            raise AttributeError("'templates' is not a folder")

    if kwargs['n_exc'] is not None:
        params_dict['spiketrains']['n_exc'] = kwargs['n_exc']
    if kwargs['n_inh'] is not None:
        params_dict['spiketrains']['n_inh'] = kwargs['n_inh']
    if kwargs['f_exc'] is not None:
        params_dict['spiketrains']['f_exc'] = kwargs['f_exc']
    if kwargs['f_inh'] is not None:
        params_dict['spiketrains']['f_inh'] = kwargs['f_inh']
    if kwargs['st_exc'] is not None:
        params_dict['spiketrains']['st_exc'] = kwargs['st_exc']
    if kwargs['st_inh'] is not None:
        params_dict['spiketrains']['st_inh'] = kwargs['st_inh']
    if kwargs['min_rate'] is not None:
        params_dict['spiketrains']['min_rate'] = kwargs['min_rate']
    if kwargs['ref_per'] is not None:
        params_dict['spiketrains']['ref_per'] = kwargs['ref_per']
    if kwargs['process'] is not None:
        params_dict['spiketrains']['process'] = kwargs['process']
    if kwargs['min_rate'] is not None:
        params_dict['spiketrains']['min_rate'] = kwargs['min_rate']
    if kwargs['duration'] is not None:
        params_dict['spiketrains']['duration'] = kwargs['duration']
    if kwargs['tstart'] is not None:
        params_dict['spiketrains']['t_start'] = kwargs['t_start']
    if kwargs['st_seed'] is not None:
        params_dict['spiketrains']['seed'] = kwargs['st_seed']

    if kwargs['min_dist'] is not None:
        params_dict['templates']['min_dist'] = kwargs['min_dist']
    if kwargs['min_amp'] is not None:
        params_dict['templates']['min_amp'] = kwargs['min_amp']
    if kwargs['max_amp'] is not None:
        params_dict['templates']['max_amp'] = kwargs['max_amp']
    if kwargs['temp_seed'] is not None:
        params_dict['templates']['seed'] = kwargs['temp_seed']
    if kwargs['overlap_thresh'] is not None:
        params_dict['templates']['overlap_threshold'] = kwargs['overlap_thresh']

    if kwargs['noise_lev'] is not None:
        params_dict['recordings']['noise_level'] = kwargs['noise_lev']
    if kwargs['modulation'] is not None:
        params_dict['recordings']['modulation'] = kwargs['modulation']

    if kwargs['chunk_noise'] is not None:
        params_dict['recordings']['chunk_noise_duration'] = kwargs['chunk_noise']
    if kwargs['chunk_filt'] is not None:
        params_dict['recordings']['chunk_filter_duration'] = kwargs['chunk_filt']
    if kwargs['no_filt'] is True:
        params_dict['recordings']['filter'] = False
    if kwargs['filt_cutoff'] is not None:
        if isinstance(kwargs['filt_cutoff'], tuple):
            kwargs['filt_cutoff'] = list(kwargs['filt_cutoff'])
        params_dict['recordings']['filter_cutoff'] = kwargs['filt_cutoff']
    if kwargs['filt_order'] is not None:
        params_dict['recordings']['filt_order'] = kwargs['filt_order']
    if kwargs['fs'] is not None:
        params_dict['recordings']['fs'] = kwargs['fs']
    else:
        params_dict['recordings']['fs'] = None
    if kwargs['sync_rate'] is not None:
        params_dict['recordings']['sync_rate'] = kwargs['sync_rate']
    else:
        params_dict['recordings']['sync_rate'] = 0
    if kwargs['sync_jitt'] is not None:
        params_dict['recordings']['sync_jitt'] = kwargs['sync_jitt']
    else:
        params_dict['recordings']['sync_jitt'] = 1
    if kwargs['noise_seed'] is not None:
        params_dict['recordings']['seed'] = kwargs['noise_seed']
    if kwargs['overlap']:
        params_dict['recordings']['overlap'] = True
    else:
        params_dict['recordings']['overlap'] = False
    if kwargs['extract_wf']:
        params_dict['recordings']['extract_wf'] = True

    if kwargs['half_dist'] is not None:
        params_dict['recordings']['half_dist'] = kwargs['half_dist']
    if kwargs['color_noise']:
        params_dict['recordings']['noise_color'] = True
    else:
        params_dict['recordings']['noise_color'] = False
    if kwargs['color_peak'] is not None:
        params_dict['recordings']['color_peak'] = kwargs['color_peak']
    if kwargs['color_q'] is not None:
        params_dict['recordings']['color_q'] = kwargs['color_q']
    if kwargs['random_noise_floor'] is not None:
        params_dict['recordings']['random_noise_floor'] = kwargs['random_noise_floor']

    if kwargs['drifting']:
        params_dict['recordings']['drifting'] = True
    else:
        params_dict['recordings']['drifting'] = False
    if kwargs['preferred_dir'] is not None and len(kwargs['preferred_dir']) == 3:
        params_dict['recordings']['preferred_dir'] = kwargs['preferred_dir']
    if kwargs['angle_tol']:
        params_dict['recordings']['angle_tol'] = kwargs['angle_tol']
    if kwargs['drift_velocity']:
        params_dict['recordings']['drift_velocity'] = kwargs['drift_velocity']
    if kwargs['t_start_drift']:
        params_dict['recordings']['t_start_drift'] = kwargs['t_start_drift']
    verbose = kwargs['verbose']

    recgen = generators.gen_recordings(templates=kwargs['templates'], params=params_dict, verbose=verbose)
    info = recgen.info

    n_neurons = info['recordings']['n_neurons']
    electrode_name = info['electrodes']['electrode_name']
    duration = info['recordings']['duration']
    noise_level = info['recordings']['noise_level']

    if kwargs['fname'] is None:
        if kwargs['drifting']:
            fname = 'recordings_%dcells_%s_%s_%.1fuV_drift_%s.h5' % (n_neurons, electrode_name, duration,
                                                               noise_level, time.strftime("%d-%m-%Y:%H:%M"))
        else:
            fname = 'recordings_%dcells_%s_%s_%.1fuV_%s.h5' % (n_neurons, electrode_name, duration,
                                                               noise_level, time.strftime("%d-%m-%Y:%H:%M"))
    else:
        fname = kwargs['fname']

    if recordings_folder is not None:
        if not os.path.isdir(recordings_folder):
            os.makedirs(recordings_folder)
    rec_path = join(recordings_folder, fname)
    save_recording_generator(recgen, rec_path)


@cli.command()
def default_config():
    """Print default configurations"""
    info, config = get_default_config()
    pprint.pprint(info)


@cli.command()
@click.argument('cell-models-folder')
def set_cell_models_folder(cell_models_folder):
    """Set default cell_models folder"""
    info, config = get_default_config()
    if os.path.isdir(cell_models_folder):
        info['cell_models_folder'] = os.path.abspath(cell_models_folder)
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default cell_models_folder to: ', cell_models_folder)
    else:
        print(cell_models_folder, ' is not a folder!')


@cli.command()
@click.argument('templates-folder')
@click.option('--create', is_flag=True, help='if True it creates the directory')
def set_templates_folder(templates_folder, create):
    """Set default templates output folder"""
    info, config = get_default_config()
    templates_folder = os.path.abspath(templates_folder)
    if os.path.isdir(templates_folder):
        info['templates_folder'] = templates_folder
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default templates_folder to: ', templates_folder)
    elif create:
        os.makedirs(templates_folder)
        info['templates_folder'] = templates_folder
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default templates_folder to: ', templates_folder)
    else:
        print(templates_folder, ' is not a folder!')


@cli.command()
@click.argument('recordings-folder')
@click.option('--create', is_flag=True, help='if True it creates the directory')
def set_recordings_folder(recordings_folder, create):
    """Set default recordings output folder"""
    info, config = get_default_config()
    recordings_folder = os.path.abspath(recordings_folder)
    if os.path.isdir(recordings_folder):
        info['recordings_folder'] = recordings_folder
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default recordings_folder to: ', recordings_folder)
    elif create:
        os.makedirs(recordings_folder)
        info['recordings_folder'] = recordings_folder
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default recordings_folder to: ', recordings_folder)
    else:
        print(recordings_folder, ' is not a folder!')


@cli.command()
@click.argument('templates-params')
def set_templates_params(templates_params):
    """Set default templates output folder"""
    info, config = get_default_config()
    templates_params = os.path.abspath(templates_params)
    if os.path.isfile(templates_params) and (templates_params.endswith('yaml') or templates_params.endswith('yml')):
        info['templates_params'] = templates_params
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default templates_params to: ', templates_params)
    else:
        print(templates_params, ' is not a yaml!')


@cli.command()
@click.argument('recordings-params')
def set_recordings_params(recordings_params):
    """Set default templates output folder"""
    info, config = get_default_config()
    recordings_params = os.path.abspath(recordings_params)
    if os.path.isfile(recordings_params) and (recordings_params.endswith('yaml') or recordings_params.endswith('yml')):
        info['recordings_params'] = recordings_params
        with open(join(config, 'mearec.conf'), 'w') as f:
            yaml.dump(info, f)
        print('Set default recordings_params to: ', recordings_params)
    else:
        print(recordings_params, ' is not a yaml!')



from setuptools import setup, find_packages

long_description = open("README.md").read()

setup(
    name="MEArec",
    version="1.2.4",
    author="Alessio Buccino",
    author_email="alessiob@ifi.uio.no",
    description="Fast and customizable simulation of extracellular recordings on Multi-Electrode-Arrays.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alejoe91/MEArec",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)  ",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy',
        'click',
        'pyyaml',
        'matplotlib',
        'neo',
        'elephant',
        'h5py',
        'MEAutility'
    ],
    entry_points={'console_scripts': 'mearec=MEArec.cli:cli'}
)